import argparse
import torch
import torch.backends.cudnn as cudnn
from torchvision import models
from NaroNet.Patch_Contrastive_Learning.data_aug.contrastive_learning_dataset import ContrastiveLearningDataset_PCL
from NaroNet.Patch_Contrastive_Learning.models.resnet_simclr import ResNetSimCLR, ResNetSimCLR_pretrained
from NaroNet.Patch_Contrastive_Learning.simclr import SimCLR
from NaroNet.Patch_Contrastive_Learning.data_aug.view_generator import ContrastiveLearningViewGenerator
from torchvision.transforms import transforms
from NaroNet.Patch_Contrastive_Learning.data_aug.gaussian_blur import GaussianBlur
import os

model_names = sorted(name for name in models.__dict__
                     if name.islower() and not name.startswith("__")
                     and callable(models.__dict__[name]))

def patch_contrastive_learning(args):
                
    # Create infer_loader
    dataset_PCL = ContrastiveLearningDataset_PCL(args, training=False)
    infer_loader = torch.utils.data.DataLoader(
        dataset_PCL, batch_size=1, shuffle=False,
        num_workers=0, pin_memory=False, drop_last=True)
    
    # Create train_loader
    dataset_PCL = ContrastiveLearningDataset_PCL(args,training=True)
    train_loader = torch.utils.data.DataLoader(
        dataset_PCL, batch_size=args['PCL_Batch_Size'], shuffle=True,
        num_workers=args['PCL_N_Workers'], pin_memory=False, drop_last=True)

    # Create ResNet model
    model = ResNetSimCLR_pretrained(in_channels=dataset_PCL.in_channels, base_model=args['PCL_CNN_Architecture'], out_dim=args['PCL_Out_Dimensions'])

    # Create optimizer
    optimizer = torch.optim.Adam(model.parameters(), args['PCL_Learning_Rate'], weight_decay=args['PCL_Weight_Decay'])

    # Create scheduler
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=len(train_loader), eta_min=0,
                                                           last_epoch=-1)

    # Train patch contrastive learning 
    simclr = SimCLR(model=model, optimizer=optimizer, scheduler=scheduler, args=args)
    if len([f for f in os.listdir(simclr.model_dir) if f.endswith('.tar')]) == 0:
        simclr.train(train_loader)

    # Create directory if it doesn't exist
    out_dir = args['path']+'Patch_Contrastive_Learning/Image_Patch_Representation/'
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    # Use trained model
    if len([f for f in os.listdir(args['path']+'Patch_Contrastive_Learning/Image_Patch_Representation/') if f.endswith('.npy')]) == 0:
        args['PCL_GPU_INDEX'] = torch.device('cpu')
        simclr.infer(infer_loader, out_dir)



import random as rand
import torch
import numpy as np
import copy
import torch
from NaroNet.utils import utilz
import torch.nn.functional as F
import math
import NaroNet.NaroNet_model.loss as loss_Op
import gc

def inductiveClustering(self,Indices,index,saveClusters,training,labels,optimizer):
    '''
    Method that loads subject's respective graph, and insert it
    in NaroNet to calculate the phenotypes and neihborhoods
        Indices: (list of ints) that specifies the indices of the subjects.
        index: (int) that specifies the index from the subjects should be obtained in this minibatch step.
        saveClusters: (boolean) if True the Clusters are saved, if False nothing happens.
        trainClustering: (boolean) if True NaroNet is trained end-to-end, if false NaroNet' clustering layers are trained in an unsupervised way.
        training: (boolean) if True NaroNet is trained, if False subjects are inferred.
        labels: (list of ints) that specifies the labels of the subjects
        optimizer: (object) with the specified NaroNet optimizer.
    '''
    
    def clustToTissueGraph(SelectedsubImIndx, chosenLabels, saveClusterPresencePheno, s_iter, saveClusterPresenceComm, total_num_nodes,training):
        for b, subIm in enumerate(SelectedsubImIndx):
            if len(self.data.y)==0:
                self.data.y = torch.tensor(chosenLabels,dtype=torch.float32).to(device=self.device)                
            saveClusterPresencePheno[subIm[1],:] += s_iter[0][b,:self.dataNOW.num_nodes[b],:].sum(-2)
            saveClusterPresenceComm[subIm[1],:] += s_iter[-1][b,:self.dataNOW.num_nodes[b],:].sum(-2)                                                                   
            total_num_nodes[subIm[1]] += self.dataNOW.num_nodes[b]      
        self.data.edge_index = self.dataNOW.edge_index
        self.data.x = self.dataNOW.x
        
        # Save Clustering of patch from a specific patch.
        if saveClusters:
            for b, subIm in enumerate(SelectedsubImIndx):
                self.dataset.saveInductiveClusters(s_iter, Indices[index+subIm[1]],subIm[0][0],b, self.args)      
        return saveClusterPresencePheno, saveClusterPresenceComm, total_num_nodes


    def subgraph_init():
        '''
        Method that initializes the subgraphs list.
        Outputs:
            subgraphs: (list of ints) that specifies the number of subgraphs a single sbject is composed by.
            chosenLabels: (list of ints) that specifies the label of each subgraph for each subject.
        '''

        # Initialize the subgraphs list and the chosenLabels
        subgraphs=[] 
        chosenLabels=[]
        
        # Iterate to obtain the number of subgraphs each subject is composed by.
        for IndicesI in Indices[index:min(index+self.args['batch_size'],len(Indices))]:             
            
            # Obtain the labels of this subject.
            chosenLabels.append(labels[IndicesI])
            
            # If True the subject has a single subgraph, if False obtain the number of subgraphs.
            if self.dataset.findLastIndex(IndicesI)==0:
                subgraphs.append([0]) 
            else:
                subgraphs.append(list(range(0,self.dataset.findLastIndex(IndicesI)))) 
        
        return subgraphs, chosenLabels
    
    def print_losses(cell_ent_loss0,cell_ent_loss1,pat_ent0,pat_ent1,args):
        if args['verbose']==1:
            print('Patient entropy (Phenotypes):'+str(pat_ent0.item())+' Neighborhoods:'+str(pat_ent1.item()))
            print('Patch entropy (Phenotypes):'+str(cell_ent_loss0.item())+' Neighborhoods:'+str(cell_ent_loss1.item()))
    
    # Initialize the subgraphs list
    subgraphs, chosenLabels = subgraph_init()

    # Initialize an empty graph
    self.data = self.dataset.generateEmptyClusteringGraph(len(subgraphs),self.args['clusters'][1]) 
    self.data.x = torch.Tensor(self.data.x).to(self.device)
    self.data.edge_index = torch.Tensor(self.data.edge_index).to(self.device)
            
    # Empty value to save the number of nodes in a slide
    total_num_nodes=np.zeros(self.args['batch_size'])
    
    # Initialize the outputs
    contrast_loss_Ph_total = []
    contrast_loss_Nb_total = []
    PatchEntropy_loss_Ph_total = []
    PatchEntropy_loss_Nb_total = []
    
    # Used to save cluster presence for visualization
    saveClusterPresencePheno = torch.zeros([len(subgraphs),int(self.args['N_Phenotypes'])])
    saveClusterPresenceComm = torch.zeros([len(subgraphs),int(self.args['N_Neighborhoods'])])


    # Initialize the optimizer 
    optimizer.zero_grad()

    # Obtain data from folder.
    self.dataNOW, subgraphs, SelectedsubImIndx=self.dataset.gatherData_parallel(self, Indices[index:index+self.args['batch_size']],subgraphs,training)
    self.dataNOW = self.dataNOW.to(self.device)
    
    # Cluster and embedding from model.
    self.dataNOW, contrast_loss_Ph, PatchEntropy_loss_Ph, contrast_loss_Nb, PatchEntropy_loss_Nb, s_iter = self.model(self.dataNOW,self.device,True,self.args)              
    
    # Unsupervised losses
    contrast_loss_Ph_total.append(contrast_loss_Ph.item()) 
    contrast_loss_Nb_total.append(contrast_loss_Nb.item())
    PatchEntropy_loss_Ph_total.append(PatchEntropy_loss_Ph.item())
    PatchEntropy_loss_Nb_total.append(PatchEntropy_loss_Nb.item())                                          

    # Assign clustering to tissue-graph
    saveClusterPresencePheno, saveClusterPresenceComm, total_num_nodes = clustToTissueGraph(SelectedsubImIndx, chosenLabels, saveClusterPresencePheno, s_iter, saveClusterPresenceComm, total_num_nodes,training)
    self.data.num_nodes = total_num_nodes
      
    return self.data, [saveClusterPresencePheno,saveClusterPresenceComm], contrast_loss_Ph, PatchEntropy_loss_Ph, contrast_loss_Nb, PatchEntropy_loss_Nb

def train(self,Indices,optimizer,training,saveClusters,labels):
    '''
    We train/test the specified subjects using NaroNet
    Indices: (list of ints) that specifies the indices of the images.
    optimizer: (object) with the specified NaroNet optimizer.
    training: (boolean) if True NaroNet is trained, if False subjects are inferred.
    trainClustering: (boolean) if True NaroNet is trained end-to-end, if false NaroNet' clustering layers are trained in an unsupervised way.
    saveClusters: (boolean) if True the Clusters are saved, if False nothing happens.
    labels: (list of ints) that specifies the labels of the subjects
    '''

    # Initialize the outputs
    total_contrast_loss_Ph = 0
    total_PatchEntropy_loss_Ph = 0 
    total_contrast_loss_Nb = 0
    total_PatchEntropy_loss_Nb = 0 
    total_contrast_loss_Ar = 0 
    total_PatchEntropy_loss_Ar = 0 
    total_Cross_entropy_loss = 0
    accuracy = 0
        
    # Shuffle Indices if we are training the model
    rand.shuffle(Indices)
    
    # Start minibatch to train/test the model
    for index in range(0,len(Indices),self.args['batch_size']):   

        # index = max(len(Indices)-self.args['batch_size'],0) if abs(index-len(Indices))<self.args['batch_size'] else index
        # print(str(index) + '-' + str(len(Indices)))
        torch.cuda.empty_cache()
        # If True NaroNet is trained, if False subjects are inferred
        if training:  
            
            # Train NaroNet
            self.model.train()

            # Obtain phenotypes and neighborhoods.            
            data, save_Inductivecluster_presence, contrast_loss_Ph, PatchEntropy_loss_Ph, contrast_loss_Nb, PatchEntropy_loss_Nb = inductiveClustering(self,Indices,index,saveClusters,training,labels,optimizer)

            # Obtain neighborhood interactions, and classify the subjects.
            out, cluster_assignment, contrast_loss_Ar, PatchEntropy_loss_Ar = self.model(data,self.device,False,self.args)                        

        else:
            
            # Subjects are inferred. 
            self.model.eval()    

            # Specify that NaroNet does not calculate the graph of gradients.
            with torch.no_grad():
                
                # Obtain phenotypes and neighborhoods.            
                data, save_Inductivecluster_presence, contrast_loss_Ph, PatchEntropy_loss_Ph, contrast_loss_Nb, PatchEntropy_loss_Nb = inductiveClustering(self,Indices,index,saveClusters,training,labels,optimizer)

                # Obtain neighborhood interactions, and classify the subjects.
                out, cluster_assignment, contrast_loss_Ar, PatchEntropy_loss_Ar = self.model(data,self.device,False,self.args)                      
        
        Cross_entropy_loss, pred_Cross_entropy, PredictedLabels_Cross_entropy = utilz.cross_entropy_loss(training, self.args, out, data.y, self.dataset, self.device, self.model)                        

        # Apply loss to update NaroNet´s parameters
        utilz.gather_apply_loss(training, contrast_loss_Ph, PatchEntropy_loss_Ph, contrast_loss_Nb, PatchEntropy_loss_Nb, contrast_loss_Ar, PatchEntropy_loss_Ar, Cross_entropy_loss, self.optimizer,self.model,self.args)    


        pred, PredictedLabels = pred_Cross_entropy, PredictedLabels_Cross_entropy
        # if self.args['NearestNeighborClassification']:
        #     pred, PredictedLabels = pred_Nearesteighbor, PredictedLabels_NearestNeighbor
        GroundTruthLabels=data.y.cpu().numpy()
        if 'PredictedLabelsAll' in locals():
            for pred_all_i in range(len(PredictedLabelsAll)):
                PredictedLabelsAll[pred_all_i] = np.concatenate((PredictedLabelsAll[pred_all_i],PredictedLabels[pred_all_i]))            
            GroundTruthLabelsAll = np.concatenate((GroundTruthLabelsAll,GroundTruthLabels))
        else:
            PredictedLabelsAll = PredictedLabels
            GroundTruthLabelsAll = GroundTruthLabels        
        for i in range(len(pred)):            
            accuracy += np.equal(pred[i],data.y[:,i].cpu().numpy()).sum()        

        # Save supervised/unsupervised loss
        total_Cross_entropy_loss += Cross_entropy_loss[0].item()
        total_contrast_loss_Ph += contrast_loss_Ph.item()
        total_PatchEntropy_loss_Ph += PatchEntropy_loss_Ph.item() 
        total_contrast_loss_Nb += contrast_loss_Nb.item()
        total_PatchEntropy_loss_Nb += PatchEntropy_loss_Nb.item() 
        total_contrast_loss_Ar += contrast_loss_Ar.item() 
        total_PatchEntropy_loss_Ar += PatchEntropy_loss_Ar.item()          
        
        # Eliminate usage of tensors in cuda.
        del data
        gc.collect()
        
        # When testing, save the clusters for visualization purposes
        if saveClusters:            
            indexes = Indices[index:min(index+self.args['batch_size'],len(Indices))]
            for idx, val in enumerate(indexes):
                self.dataset.save_cluster_and_attention(idx, val, save_Inductivecluster_presence, cluster_assignment)

    n_iter = len(list(range(0,len(Indices),self.args['batch_size'])))

    return GroundTruthLabelsAll, PredictedLabelsAll, (accuracy / PredictedLabelsAll[0].shape[0])/len(pred), total_Cross_entropy_loss/n_iter, total_contrast_loss_Ph/n_iter, total_PatchEntropy_loss_Ph/n_iter, total_contrast_loss_Nb/n_iter, total_PatchEntropy_loss_Nb/n_iter, total_contrast_loss_Ar/n_iter, total_PatchEntropy_loss_Ar/n_iter, Indices



import numpy as np
import torch
from torch import nn
from torchvision.transforms import transforms

np.random.seed(0)


class GaussianBlur(object):
    """blur a single image on CPU"""
    def __init__(self, kernel_size, in_channels):
        radias = kernel_size // 2
        kernel_size = radias * 2 + 1
        self.in_channels = in_channels
        self.blur_h = nn.Conv2d(in_channels, in_channels, kernel_size=(kernel_size, 1),
                                stride=1, padding=0, bias=False, groups=in_channels)
        self.blur_v = nn.Conv2d(in_channels, in_channels, kernel_size=(1, kernel_size),
                                stride=1, padding=0, bias=False, groups=in_channels)
        self.k = kernel_size
        self.r = radias

        self.blur = nn.Sequential(
            nn.ReflectionPad2d(radias),
            self.blur_h,
            self.blur_v
        )

        self.pil_to_tensor = transforms.ToTensor()
        self.tensor_to_pil = transforms.ToPILImage()

    def __call__(self, img):
        # img = self.pil_to_tensor(img).unsqueeze(0)

        sigma = np.random.uniform(0.1, 2.0)
        x = np.arange(-self.r, self.r + 1)
        x = np.exp(-np.power(x, 2) / (2 * sigma * sigma))
        x = x / x.sum()
        x = torch.from_numpy(x).view(1, -1).repeat(self.in_channels, 1)

        self.blur_h.weight.data.copy_(x.view(self.in_channels, 1, self.k, 1))
        self.blur_v.weight.data.copy_(x.view(self.in_channels, 1, 1, self.k))

        with torch.no_grad():
            img = self.blur(img)
            img = img.squeeze()

        return img


import torch
import torch.nn.functional as F
import NaroNet.NaroNet_model.pooling as pooling
import NaroNet.NaroNet_model.loss as loss
import NaroNet.NaroNet_model.GNN as GNN
import NaroNet.utils.utilz
import numpy as np

class NaroNet_model(torch.nn.Module):
    def __init__(self, num_features, labels, hidden, num_nodes, clusts, args):
        super(NaroNet_model, self).__init__()
        # Initialization
        if  args['LSTM']:
            self.features = hidden
        else:
            self.features = num_features        
        self.args = args         
        self.hidden = hidden                               
        
        # Phenotypes             
        if args['LSTM']:
            self.LSTM_phenoClust = GNN.LSTM(num_features, hidden, args)                
        if args['GLORE']:                
            self.GloRe_phenoClust = GNN.GloRe_Unit(self.features, hidden, normalize=True)
                
        # FCN  
        self.pheno_blocks_clust = GNN.phenoNN8(self.features, hidden,clusts[0], args, mode='Multiplication')
                        
        # First Graph Neural Network Embedding 
        if args['LSTM']:
            self.LSTM_GNN1_Clust = GNN.LSTM(num_features, hidden, args)
            self.LSTM_GNN1_Emb = GNN.LSTM(num_features, hidden, args)
        if args['GLORE']:
            self.GloRe_GNN1_Clust = GNN.GloRe_Unit(self.features, hidden, normalize=True)
            self.GloRe_GNN1_Emb = GNN.GloRe_Unit(self.features, hidden, normalize=True)
        if args['modeltype']=='SAGE':
            self.embed_pool_block_emb = GNN.GNN(self.features, hidden, hidden, args, args['n-hops'], mode='SparseMultiplication')
            self.embed_pool_block_clust = GNN.GNN(self.features, hidden,  clusts[1], args,args['n-hops'], mode='SparseMultiplication')               
            
        # Second Graph Neural Network Embedding
        if args['modeltype']=='SAGE':
            self.embed_pool_block2_clust = GNN.GNN(hidden, hidden,  clusts[2], args,1, mode='Multiplication')
        
        # Unsupervised learning
        if args['UnsupContrast']:
            self.lin1_unsupA = torch.nn.Linear(clusts[0],hidden+hidden)
            self.lin1_unsupB = torch.nn.Linear(hidden+hidden,hidden)
            self.lin2_unsupA = torch.nn.Linear(clusts[1],hidden+hidden)
            self.lin2_unsupB = torch.nn.Linear(hidden+hidden,hidden)
            self.lin3_unsupA = torch.nn.Linear(clusts[2],hidden+hidden)
            self.lin3_unsupB = torch.nn.Linear(hidden+hidden,hidden)
            self.lr_unsup = torch.nn.Parameter(torch.ones(1))

        # Fully connected Layer
        self.lin1_1 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], hidden+hidden+hidden+hidden, bias=True)          
        self.BNLast_1 = torch.nn.BatchNorm1d(hidden+hidden+hidden+hidden,track_running_stats=False)
        self.lin2_1 = torch.nn.Linear(hidden+hidden+hidden+hidden, int(max(labels[0])) if len(labels[0])>5 else len(labels[0]), bias=True)
        if len(args['experiment_Label'])>1:
            self.lin1_2 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], hidden+hidden+hidden+hidden, bias=True)             
            self.BNLast_2 = torch.nn.BatchNorm1d(hidden+hidden+hidden+hidden,track_running_stats=False)
            self.lin2_2 = torch.nn.Linear(hidden+hidden+hidden+hidden, int(max(labels[1])) if len(labels[1])>5 else len(labels[1]), bias=True)
        if len(args['experiment_Label'])>2:
            self.lin1_3 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], hidden+hidden+hidden+hidden, bias=True)             
            self.BNLast_3 = torch.nn.BatchNorm1d(hidden+hidden+hidden+hidden,track_running_stats=False)
            self.lin2_3 = torch.nn.Linear(hidden+hidden+hidden+hidden, int(max(labels[2])) if len(labels[2])>5 else len(labels[2]), bias=True)
        if len(args['experiment_Label'])>3:
            self.lin1_4 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], hidden+hidden+hidden+hidden, bias=True)                                 
            self.BNLast_4 = torch.nn.BatchNorm1d(hidden+hidden+hidden+hidden,track_running_stats=False)        
            self.lin2_4 = torch.nn.Linear(hidden+hidden+hidden+hidden, int(max(labels[3])) if len(labels[3])>5 else len(labels[3]), bias=True)

    def reset_parameters(self):        
                
        # Phenotypes          
        if self.args['LSTM']:
            for name, module in self.LSTM_phenoClust.named_children():
                module.reset_parameters()                
        if self.args['GLORE']:
            self.GloRe_phenoClust.reset_parameters()                                            
        if self.args['DeepSimple']:
            for pheno in self.pheno_blocks_clust:
                pheno.reset_parameters()
        else:
            self.pheno_blocks_clust.reset_parameters()    
            
        # First GNN        
        if self.args['LSTM']:
            for name, module in self.LSTM_GNN1_Clust.named_children():
                module.reset_parameters()
            for name, module in self.LSTM_GNN1_Emb.named_children():
                module.reset_parameters()  
        if self.args['GLORE']:
            self.GloRe_GNN1_Emb.reset_parameters()                
            self.GloRe_GNN1_Clust.reset_parameters()
        self.embed_pool_block_emb.reset_parameters()      
        self.embed_pool_block_clust.reset_parameters()                  
        
        # Second GNN        
        self.embed_pool_block2_clust.reset_parameters()              

        # # SAGPOOL
        # self.X1_SAGPOOL.reset_parameters()
        # self.X2_SAGPOOL.reset_parameters()
        
        # SUP CON LOSS
        # self.SUPCONlin.reset_parameters()

        # Unsupervised learning
        if self.args['UnsupContrast']:
            self.lin1_unsupA.reset_parameters()
            self.lin1_unsupB.reset_parameters()
            self.lin2_unsupA.reset_parameters()
            self.lin2_unsupB.reset_parameters()
            self.lin3_unsupA.reset_parameters()
            self.lin3_unsupB.reset_parameters()
            self.lr_unsup = torch.nn.Parameter(torch.ones(1))

        if self.args['ortho']:
            self.lr_ortho = torch.nn.Parameter(torch.ones(1))

        if self.args['orthoColor']:
            self.lr_orthoColor = torch.nn.Parameter(torch.ones(1))

        if self.args['F-test']:
            self.ftest_lin.reset_parameters()
            self.ftest_BNLast = torch.nn.BatchNorm1d(self.hidden+self.hidden,track_running_stats=False)
            self.ftest_lin2.reset_parameters()


        # Fully connected Layer               
        self.lin1_1.reset_parameters()
        self.BNLast_1 = torch.nn.BatchNorm1d(self.hidden+self.hidden+self.hidden+self.hidden,track_running_stats=False)
        self.lin2_1.reset_parameters()
        if len(self.args['experiment_Label'])>1:
            self.lin1_2.reset_parameters()           
            self.BNLast_2 = torch.nn.BatchNorm1d(self.hidden+self.hidden+self.hidden+self.hidden,track_running_stats=False)
            self.lin2_2.reset_parameters()
        if len(self.args['experiment_Label'])>2:
            self.lin1_3.reset_parameters()
            self.BNLast_3 = torch.nn.BatchNorm1d(self.hidden+self.hidden+self.hidden+self.hidden,track_running_stats=False)
            self.lin2_3.reset_parameters()
        if len(self.args['experiment_Label'])>3:
            self.lin1_4.reset_parameters()
            self.BNLast_4 = torch.nn.BatchNorm1d(self.hidden+self.hidden+self.hidden+self.hidden,track_running_stats=False)
            self.lin2_4.reset_parameters()
                    
    def softmaxToClst(self,s,args,device):
        s=F.softmax(s,dim=-1)
        alpha=0.1
        if args['1cell1cluster']:
            # s = torch.where(s>=s.max(-1).values.unsqueeze(-1).repeat(1,1,s.shape[2]), alpha*s+(1-alpha), torch.tensor([0],dtype=torch.float32).to(device))
            s = torch.where(s>=s.max(-1).values.unsqueeze(-1).repeat(1,1,s.shape[2]), s, s)
        s = torch.where(s>=args['attntnThreshold'], s, s*0)
        # s[s<args['attntnThreshold']]=s[s<args['attntnThreshold']]*0.1
        # Gamma Correction: 0.5
        # s = s**1.5
        # s = torch.where(s>3*args['attntnThreshold']/s.shape[2], s, torch.tensor([0],dtype=torch.float32).to(device))
        # print('Cluster Confidence:', s.max(-1).values.mean())
        # print('Activated Nodes:', (s.max(-1)[0]>0).sum().detach().cpu().numpy()/(s.shape[0]*s.shape[1]))
        return s#/s.sum(dim=-1,keepdim=True)

    def sigmoidToAttn(self,s,args,device):
        s=torch.sigmoid(s)
        # s=F.softmax(s,dim=-1)
        if args['1cell1cluster']:
            s = torch.where(s>=s.max(-1).values.unsqueeze(-1).repeat(1,1,s.shape[2]), s, torch.tensor([0],dtype=torch.float32).to(device))        
        s = torch.where(s>=args['attntnThreshold'], s, s*0.001)
        # s[-1][s[-1]<args['attntnThreshold']]=s[-1][s[-1]<args['attntnThreshold']]*0.1
        # s = torch.where(s>=s.max(-1).values.unsqueeze(-1).repeat(1,1,s.shape[2]), s, torch.tensor([0],dtype=torch.float32).to(device))
        # s = torch.where(s>args['attntnThreshold'], s, torch.tensor([0],dtype=torch.float32).to(device))
        return s#/s.sum(dim=-1,keepdim=True)
    
    def poolingToClst(self,s,device,num_nodes):
        Spatient = torch.zeros(s.shape[0],s.shape[-1],dtype=torch.float32).to(device)
        for i in range(s.shape[0]):
            Spatient[i,:] = (s[i,:num_nodes[i],:].sum(0)/num_nodes[i])   
        return Spatient

    def SAGPOOL(self,score,x,num_nodes):
        score=F.softmax(score,dim=1)        
        for b in range(x.shape[0]):
            s,idx = score[b,:,0].sort(-1,descending=True)
            x[b:,:int(num_nodes),:] = score[b,idx[:int(num_nodes)],:]*x[b,idx[:int(num_nodes)],:]
        return x[:,:int(num_nodes),:]

    def ObtainPhenotypesClustering(self, x, data, device, args):                        
        if args['DeepSimple']: # Multiple Neural Networks but they have few parameters...
            self.s=[]
            for pheno in self.pheno_blocks_clust:
                self.s.append(pheno(x, data.edge_index, device, data.num_nodes,args))
            self.s = torch.stack(self.s,dim=-1)
            self.s = [torch.squeeze(self.s,dim=-2)]
        else:
            self.s=[self.pheno_blocks_clust(x, data.edge_index, device, data.num_nodes,args)]        
        return self.s
    
    def drop_learning_modules(self, S, args):
        if not args['Phenotype_Learning']:
            S[0] = S[0]*0
        if not args['Neighborhood_Learning']:
            S[1] = S[1]*0
        if not args['Area_Learning']:
            S[2] = S[2]*0
        return S

    def ClassifyPatients(self, args):
        x = []

        self.S = self.drop_learning_modules(self.S, args)
        
        # Fully connected Layer               
        x.append(F.relu(self.lin1_1(torch.cat(self.S,dim=-1))))
        if args['Batch_Normalization']:
            x[0] = self.BNLast_1(x[0]) if x[0].shape[0]>1 else x[0]                        
        x[0] = F.dropout(x[0], p=args['dropoutRate'], training=self.training)
        x[0] = self.lin2_1(x[0])          
        if len(self.args['experiment_Label'])>1:
            x.append(F.relu(self.lin1_2(torch.cat(self.S,dim=-1))))
            if args['Batch_Normalization']:
                x[1] = self.BNLast_2(x[1]) if x[1].shape[0]>1 else x[1]                        
            x[1] = F.dropout(x[1], p=args['dropoutRate'], training=self.training)
            x[1] = self.lin2_2(x[1])  
        if len(self.args['experiment_Label'])>2:
            x.append(F.relu(self.lin1_3(torch.cat(self.S,dim=-1))))
            if args['Batch_Normalization']:
                x[2] = self.BNLast_3(x[2]) if x[2].shape[0]>1 else x[2]                        
            x[2] = F.dropout(x[2], p=args['dropoutRate'], training=self.training)
            x[2] = self.lin2_3(x[2])  
        if len(self.args['experiment_Label'])>3:
            x.append(F.relu(self.lin1_4(torch.cat(self.S,dim=-1))))
            if args['Batch_Normalization']:
                x[3] = self.BNLast_4(x[3]) if x[3].shape[0]>1 else x[3]                        
            x[3] = F.dropout(x[3], p=args['dropoutRate'], training=self.training)
            x[3] = self.lin2_4(x[3])  
                
        return x

    def TissueCommunitiesInter_forward(self, data, doClustering, device, args):
        
        # GNN Forward to obtain clusters.
        self.s.append(self.embed_pool_block2_clust(data.x, data.edge_index, device, data.num_nodes,args))                            
        
        # Obtain Cell Entropy Loss
        ortho_color, pearsonCoeffSUP, pearsonCoeffUNSUP, ortho, cell_ent_loss = loss.ortho_and_mincut_loss(data,F.softmax(self.s[-1],dim=2), args, np.ones(self.s[-1].shape[0],dtype=int)*self.s[-1].shape[1],device)  if args['ClusteringOrAttention'] else loss.ortho_and_mincut_loss(data,torch.sigmoid(self.s[-1]),args, np.ones(self.s[-1].shape[0],dtype=int)*self.s[-1].shape[1],device)                    

        # Apply Softmax to cluster assignment             
        self.s[-1] = self.softmaxToClst(self.s[-1],args,device) if args['ClusteringOrAttention'] else self.sigmoidToAttn(self.s[-1],args,device)        
        
        # Node-Pooling 
        data.x, data.edge_index, minCut = pooling.Sparse_Pooling(data.y,data.x, data.edge_index, self.s[-1], device, args,self)  # Regions                      
        
        # Obtain Patient Concentration for each cluster
        self.S.append(self.poolingToClst(self.s[-1],device,[self.s[-1].shape[1]]*self.s[-1].shape[0]))                        
        
        # Obtain Patient Entropy Loss
        pat_ent = loss.pat_loss(self.S[-1],args, data.num_nodes,device)      

        self.s_interaction.append(data.edge_index) # Extract interactions between 2nd order Tissue-communities
        self.XS.append(data.x.max(-1)[0]) # Extract activations from 2nd order Tissue-communities
        self.s[-1] = self.s[-1].to('cpu') # Eliminate cluser assignment from GPU
        return data.x, data.edge_index, ortho_color, pearsonCoeffSUP, pearsonCoeffUNSUP, minCut, ortho, cell_ent_loss, pat_ent

    def TissueCommunities_forward(self, data, doClustering, device, args):
        if args['LSTM']:
            self.s.append(self.embed_pool_block_clust(self.LSTM_GNN1_Clust(data.x,device), data.edge_index, device, data.num_nodes,args))
            data.x = self.embed_pool_block_emb(self.LSTM_GNN1_Emb(data.x,device), data.edge_index, device, data.num_nodes,args)                      
        else:
            if args['GLORE']:    
                self.s.append(self.embed_pool_block_clust(self.GloRe_GNN1_Clust(data.x), data.edge_index, device, data.num_nodes,args))                    
                data.x = self.embed_pool_block_emb(self.GloRe_GNN1_Clust(data.x), data.edge_index, device, data.num_nodes,args)                                          
            else:
                self.s.append(self.embed_pool_block_clust(data.x, data.edge_index, device, data.num_nodes,args))                    
                data.x = self.embed_pool_block_emb(data.x, data.edge_index, device, data.num_nodes,args)                      
        
        # Obtain Cell Entropy Loss
        ortho_color1, pearsonCoeffSUP1, pearsonCoeffUnsup1, ortho1, cell_ent_loss1 = loss.ortho_and_mincut_loss(data,F.softmax(self.s[-1],dim=2),args, data.num_nodes,device) if args['ClusteringOrAttention'] else loss.ortho_and_mincut_loss(data,torch.sigmoid(self.s[-1]),args, data.num_nodes,device)
                
        # Apply Softmax to cluster assignment       
        self.s[-1] = self.softmaxToClst(self.s[-1],args,device) if args['ClusteringOrAttention'] else self.sigmoidToAttn(self.s[-1],args,device)   

        # Node-Pooling     
        data.x, data.edge_index, minCut1 = pooling.Dense_Pooling(data.y,data.x, args['hiddens'], data.edge_index, self.s[-1], device, data.num_nodes,doClustering, args,self, True)  # Tissue-Communities                                             
        
        # Obtain Patient Concentration for each cluster
        self.S.append(self.poolingToClst(self.s[-1],device,data.num_nodes))            
        
        # Obtain Patient Entropy Loss
        pat_ent1 = loss.pat_loss(self.S[-1],args, data.num_nodes,device)      

        self.s_interaction.append(data.edge_index)
        self.XS.append(data.x.max(-1)[0])
        
        # self.NN_loss += -self.s[-1].max(-1).values.mean()
        self.s[-1] = self.s[-1].to('cpu')
        return data.x, data.edge_index, ortho_color1, pearsonCoeffSUP1, pearsonCoeffUnsup1, minCut1, ortho1, cell_ent_loss1,pat_ent1

    def Phenotypes_forward(self, data, doClustering, device, args):
        if args['LSTM']:
            self.s, self.S = self.ObtainPhenotypesClustering(data.x,data, device, args)                        
        else:
            if args['GLORE']:    
                self.s = self.ObtainPhenotypesClustering(self.GloRe_phenoClust(data.x),data, device, args)                                           
            else:
                self.s = self.ObtainPhenotypesClustering(data.x,data, device, args)                           
               

        # Obtain Cell Entropy Loss
        ortho_color0, pearsonCoeffSUP0, pearsonCoeffUnsup0, ortho0, cell_ent_loss0 = loss.ortho_and_mincut_loss(data,F.softmax(self.s[0],dim=2),args, data.num_nodes,device) if args['ClusteringOrAttention'] else loss.ortho_and_mincut_loss(data,torch.sigmoid(self.s[0]),args, data.num_nodes,device)
                        
        # Apply Softmax to cluster assignment
        self.s[-1] = self.softmaxToClst(self.s[-1],args,device) if args['ClusteringOrAttention'] else self.sigmoidToAttn(self.s[-1],args,device)                        
        
        # Calculate Clusters Ifñormatioñ
        self.XS, self.s_interaction, minCut0 = pooling.Dense_Pooling(data.y,data.x, args['hiddens'], data.edge_index, self.s[0], device, data.num_nodes,doClustering, args, self, False)  # Phenotyep                                                                                    
        
        # Obtain Patient Concentration for each cluster
        self.S = [self.poolingToClst(self.s[-1],device,data.num_nodes)]
        
        # Obtain Patient Entropy Loss
        pat_ent0 = loss.pat_loss(self.S[-1],args, data.num_nodes,device)      

        self.s_interaction = [self.s_interaction.to('cpu')]
        # self.NN_loss = -self.s[0].max(-1).values.mean()
        self.s[0] = self.s[0].to('cpu')
    # self.X = self.X1_SAGPOOL(self.X, data.edge_index, device, data.num_nodes,args)
    # a = self.X1_SAGPOOL(data.x, data.edge_index, device, data.num_nodes,args)
    # self.X = [self.SAGPOOL(a,data.x,args['clusters1']).max(-1)[0]]
    # self.X = [self.X.max(-2)[0]]
    # self.X[0] =self.X1_SAGPOOL(self.X[0], data.edge_index, device, data.num_nodes,args)[:,:int(args['clusters1']),0]
        self.XS= [self.XS.max(-1)[0]]
        return ortho_color0, pearsonCoeffSUP0, pearsonCoeffUnsup0, minCut0, ortho0, cell_ent_loss0,pat_ent0

    def forward(self, data, device, saveClusters, trainClustering, doClustering,index,Indices,labels,args):
        
        if doClustering:                                
            # Obtain Phenotypes
            ortho_color0, pearsonCoeffSUP0, pearsonCoeffUnsup0, minCut0, ortho0, cell_ent_loss0, pat_ent0 = self.Phenotypes_forward(data, doClustering, device, args)
            
            # Obtain Tissue-communities                        
            data.x, data.edge_index, ortho_color1, pearsonCoeffSUP1, pearsonCoeffUnsup1, minCut1, ortho1, cell_ent_loss1, pat_ent1 = self.TissueCommunities_forward(data, doClustering, device, args)
            
            # Return Phenotypes and Tissue-Communities            
            return data, ortho_color0, ortho_color1, minCut0, minCut1, ortho0, ortho1, cell_ent_loss0, cell_ent_loss1, pat_ent0, pat_ent1, pearsonCoeffSUP0+pearsonCoeffSUP1, pearsonCoeffUnsup0+pearsonCoeffUnsup1, self.s                        
        
        else:        
            # Obtain Tissue-Communities Interactions
           data.x, data.edge_index, ortho_color2, pearsonCoeffSUP2, pearsonCoeffUNSUP2, minCut2, ortho2,  cell_ent_loss2, pat_ent2 = self.TissueCommunitiesInter_forward(data, doClustering, device, args)

        # Classify patients using Patient Embedding
        if not doClustering:             
            x = self.ClassifyPatients(args)                       
        else:
            x=torch.Tensor([0,0,0]).to(device)

        # F-test to obtain the least number possible of clusters.
        # f_test_loss = loss.f_test_loss(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1),data.y,device,self.lin1,self.lin2,self.BNLast,args)
        f_test_loss = 0
        
        # Maximize patient class embeddings
        # supconloss = loss.SupConLoss_Total(args,self.S,labels,data,device)
        supconloss = 0
        
        # Unsupervised Clustering
        unsupconloss = loss.UnsupConLoss_Total(args, self.S, data, device, self)
                
        return x, supconloss, minCut2, ortho2, self.s[2:], self.s_interaction, 0, [], pearsonCoeffSUP2, pearsonCoeffUNSUP2, 0, f_test_loss, unsupconloss, pat_ent2, cell_ent_loss2 


import matplotlib.pyplot as plt
import numpy as np

def plot_roc(fpr,tpr,roc_auc):
    plt.figure()
    lw = 3
    plt.plot(fpr, tpr, color='darkorange',
            lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([-0.02, 1.02])
    plt.ylim([-0.02, 1.02])
    plt.xlabel('False Positive Rate',fontsize=16)
    plt.ylabel('True Positive Rate',fontsize=16)
    plt.title('Receiver operating characteristic',fontsize=16)
    plt.legend(loc="lower right",fontsize=16)    
    return plt

def plot_prec_recall(recall,precision,AP,f1_score):
    plt.figure()
    lw = 3
    plt.plot(recall, precision, color='navy',
            lw=lw, label='P-R curve (AP = %0.2f)' % AP)    
    plt.xlim([-0.02, 1.02])
    plt.ylim([-0.02, 1.02])
    plt.xlabel('Recall',fontsize=16)
    plt.ylabel('Precision',fontsize=16)
    plt.title('Precision-Recall curve (F1-score: {:.2f})'.format(f1_score),fontsize=16)
    plt.legend(loc="lower left",fontsize=16)    
    return plt

def plot_training_loss_acc(train_info=[],val_info=[], title=[], label=[], ylabel=[]):    
    train_info_mean = [np.mean(t) for t in train_info]
    train_info_std = [np.std(t) for t in train_info]
    val_info_mean = [np.mean(t) for t in val_info]
    val_info_std = [np.std(t) for t in val_info]
    plt.errorbar(np.linspace(1, len(train_info_mean), len(train_info_mean),endpoint=True),train_info_mean,yerr=train_info_std, label=label[0])
    plt.errorbar(np.linspace(1, len(val_info_mean), len(val_info_mean)),val_info_mean,yerr=val_info_std, label=label[1])
    plt.title(title, fontsize=16)
    plt.xlabel('Epochs',fontsize=16)
    plt.ylabel(ylabel, fontsize=16)
    plt.legend(fontsize='large')
    return plt




import numpy as np
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler
from sklearn_pandas import DataFrameMapper

import torch
import torchtuples as tt

from pycox.datasets import metabric
from NaroNet.pycox_code.pycox.models import CoxTime
from NaroNet.pycox_code.pycox.models.cox_time import MLPVanillaCoxTime
from NaroNet.pycox_code.pycox.evaluation import EvalSurv

df_train = metabric.read_df()
df_test = df_train.sample(frac=0.2)
df_train = df_train.drop(df_test.index)
df_val = df_train.sample(frac=0.2)
df_train = df_train.drop(df_val.index)
df_train.head()

# Covariates
cols_standardize = ['x0', 'x1', 'x2', 'x3', 'x8']
cols_leave = ['x4', 'x5', 'x6', 'x7']
standardize = [([col], StandardScaler()) for col in cols_standardize]
leave = [(col, None) for col in cols_leave]
x_mapper = DataFrameMapper(standardize + leave)
x_train = x_mapper.fit_transform(df_train).astype('float32')
x_val = x_mapper.transform(df_val).astype('float32')
x_test = x_mapper.transform(df_test).astype('float32')

# Labels
labtrans = CoxTime.label_transform()
get_target = lambda df: (df['duration'].values, df['event'].values)
y_train = labtrans.fit_transform(*get_target(df_train))
y_val = labtrans.transform(*get_target(df_val))
durations_test, events_test = get_target(df_test)
val = tt.tuplefy(x_val, y_val)

# Neural Network
in_features = x_train.shape[1]
num_nodes = [32, 32]
batch_norm = True
dropout = 0.1
net = MLPVanillaCoxTime(in_features, num_nodes, batch_norm, dropout)
model = CoxTime(net, tt.optim.Adam, labtrans=labtrans)
model.optimizer.set_lr(0.01)

# Train model
epochs = 512
batch_size = 256
callbacks = [tt.callbacks.EarlyStopping()]
verbose = True
log = model.fit(x_train, y_train, batch_size, epochs, callbacks, verbose,val_data=val.repeat(10).cat())


# Predict
_ = model.compute_baseline_hazards()
surv = model.predict_surv_df(x_test)



import warnings

import matplotlib.pyplot as plt
from matplotlib import lines
import matplotlib.transforms as mtransforms
from matplotlib.font_manager import FontProperties
import numpy as np
import pandas as pd
import seaborn as sns
from seaborn.utils import remove_na
from scipy import stats

DEFAULT = object()
def raise_expected_got(expected, for_, got, error_type=ValueError):
    """Raise a standardized error message.

    Raise an `error_type` error with the message
        Expected `expected` for `for_`; got `got` instead.
    Or, if `for_` is `None`,
        Expected `expected`; got `got` instead.

    """
    if for_ is not None:
        raise error_type(
            'Expected {} for {}; got {} instead.'.format(expected, for_, got)
        )
    else:
        raise error_type(
            'Expected {}; got {} instead.'.format(expected, got)
        )


def assert_is_in(x, valid_values, error_type=ValueError, label=None):
    """Raise an error if x is not in valid_values."""
    if x not in valid_values:
            raise_expected_got('one of {}'.format(valid_values), label, x)


class StatResult:
    def __init__(self, test_str, test_short_name, stat_str, stat, pval):
        self.test_str = test_str
        self.test_short_name = test_short_name
        self.stat_str = stat_str
        self.stat = stat
        self.pval = pval

    @property
    def formatted_output(self):
        if self.stat_str is None and self.stat is None:
            stat_summary = '{}, P_val:{:.3e}'.format(self.test_str, self.pval)
        else:
            stat_summary = '{}, P_val={:.3e} {}={:.3e}'.format(
                self.test_str, self.pval, self.stat_str, self.stat
            )
        return stat_summary

    def __str__(self):
        return self.formatted_output

def stat_test(
    box_data1,
    box_data2,
    test,
    comparisons_correction=None,
    num_comparisons=1,
    **stats_params
):
    """Get formatted result of two sample statistical test.

    Arguments
    ---------
    bbox_data1, bbox_data2
    test: str
        Statistical test to run. Must be one of:
        - `Levene`
        - `Mann-Whitney`
        - `Mann-Whitney-gt`
        - `Mann-Whitney-ls`
        - `t-test_ind`
        - `t-test_welch`
        - `t-test_paired`
        - `Wilcoxon`
        - `Kruskal`
    comparisons_correction: str or None, default None
        Method to use for multiple comparisons correction. Currently only the
        Bonferroni correction is implemented.
    num_comparisons: int, default 1
        Number of comparisons to use for multiple comparisons correction.
    stats_params
        Additional keyword arguments to pass to scipy stats functions.

    Returns
    -------
    StatResult object with formatted result of test.

    """
    # Check arguments.
    assert_is_in(
        comparisons_correction,
        ['bonferroni', None],
        label='argument `comparisons_correction`',
    )

    # Switch to run scipy.stats hypothesis test.
    if test == 'Levene':
        stat, pval = stats.levene(box_data1, box_data2, **stats_params)
        result = StatResult(
            'Levene test of variance', 'levene', 'stat', stat, pval
        )
    elif test == 'Mann-Whitney':
        if len(box_data1)==len(box_data2):
            if (np.array(box_data1)==np.array(box_data2)).sum()<len(box_data1):
                u_stat, pval = stats.mannwhitneyu(box_data1, box_data2, alternative='two-sided', **stats_params)
            else:
                u_stat=0
                pval=1
        else:
            u_stat, pval = stats.mannwhitneyu(box_data1, box_data2, alternative='two-sided', **stats_params)
        result = StatResult(
            'Mann-Whitney-Wilcoxon test two-sided',
            'M.W.W.',
            'U_stat',
            u_stat,
            pval,
        )
    elif test == 'Mann-Whitney-gt':
        u_stat, pval = stats.mannwhitneyu(
            box_data1, box_data2, alternative='greater', **stats_params
        )
        result = StatResult(
            'Mann-Whitney-Wilcoxon test greater',
            'M.W.W.',
            'U_stat',
            u_stat,
            pval,
        )
    elif test == 'Mann-Whitney-ls':
        u_stat, pval = stats.mannwhitneyu(
            box_data1, box_data2, alternative='less', **stats_params
        )
        result = StatResult(
            'Mann-Whitney-Wilcoxon test smaller',
            'M.W.W.',
            'U_stat',
            u_stat,
            pval,
        )
    elif test == 't-test_ind':
        stat, pval = stats.ttest_ind(a=box_data1, b=box_data2, **stats_params)
        result = StatResult(
            't-test independent samples', 't-test_ind', 'stat', stat, pval
        )
    elif test == 't-test_welch':
        stat, pval = stats.ttest_ind(
            a=box_data1, b=box_data2, equal_var=False, **stats_params
        )
        result = StatResult(
            'Welch\'s t-test independent samples',
            't-test_welch',
            'stat',
            stat,
            pval,
        )
    elif test == 't-test_paired':
        stat, pval = stats.ttest_rel(a=box_data1, b=box_data2, **stats_params)
        result = StatResult(
            't-test paired samples', 't-test_rel', 'stat', stat, pval
        )
    elif test == 'Wilcoxon':
        zero_method_default = len(box_data1) <= 20 and "pratt" or "wilcox"
        zero_method = stats_params.get('zero_method', zero_method_default)
        print("Using zero_method ", zero_method)
        stat, pval = stats.wilcoxon(
            box_data1, box_data2, zero_method=zero_method, **stats_params
        )
        result = StatResult(
            'Wilcoxon test (paired samples)', 'Wilcoxon', 'stat', stat, pval
        )
    elif test == 'Kruskal':
        stat, pval = stats.kruskal(box_data1, box_data2, **stats_params)
        test_short_name = 'Kruskal'
        result = StatResult(
            'Kruskal-Wallis paired samples', 'Kruskal', 'stat', stat, pval
        )
    else:
        result = StatResult(None, '', None, None, np.nan)

    # Optionally, run multiple comparisons correction.
    if comparisons_correction == 'bonferroni':
        result.pval = bonferroni(result.pval, num_comparisons)
        result.test_str = result.test_str + ' with Bonferroni correction'
    elif comparisons_correction is None:
        pass
    else:
        # This should never be reached because `comparisons_correction` must
        # be a valid correction method or None.
        raise RuntimeError('Unexpectedly reached end of switch.')

    return result


def bonferroni(p_values, num_comparisons='auto'):
    """Apply Bonferroni correction for multiple comparisons.

    The Bonferroni correction is defined as
        p_corrected = min(num_comparisons * p, 1.0).

    Arguments
    ---------
    p_values: scalar or list-like
        One or more p_values to correct.
    num_comparisons: int or `auto`
        Number of comparisons. Use `auto` to infer the number of comparisons
        from the length of the `p_values` list.

    Returns
    -------
    Scalar or numpy array of corrected p-values.

    """
    # Input checks.
    if np.ndim(p_values) > 1:
        raise_expected_got(
            'Scalar or list-like', 'argument `p_values`', p_values
        )
    if num_comparisons != 'auto':
        try:
            # Raise a TypeError if num_comparisons is not numeric, and raise
            # an AssertionError if it isn't int-like.
            assert np.ceil(num_comparisons) == num_comparisons
        except (AssertionError, TypeError) as e:
            raise_expected_got(
                'Int or `auto`', 'argument `num_comparisons`', num_comparisons
            )

    # Coerce p_values to numpy array.
    p_values_array = np.atleast_1d(p_values)

    if num_comparisons == 'auto':
        # Infer number of comparisons
        num_comparisons = len(p_values_array)
    elif len(p_values_array) > 1 and num_comparisons != len(p_values_array):
        # Warn if multiple p_values have been passed and num_comparisons is
        # set manually.
        warnings.warn(
            'Manually-specified `num_comparisons={}` differs from number of '
            'p_values to correct ({}).'.format(
                num_comparisons, len(p_values_array)
            )
        )

    # Apply correction by multiplying p_values and thresholding at p=1.0
    p_values_array *= num_comparisons
    p_values_array = np.min(
        [p_values_array, np.ones_like(p_values_array)], axis=0
    )

    if len(p_values_array) == 1:
        # Return a scalar if input was a scalar.
        return p_values_array[0]
    else:
        return p_values_array



def pval_annotation_text(x, pvalue_thresholds):
    single_value = False
    if type(x) is np.array:
        x1 = x
    else:
        x1 = np.array([x])
        single_value = True
    # Sort the threshold array
    pvalue_thresholds = pd.DataFrame(pvalue_thresholds).sort_values(by=0, ascending=False).values
    x_annot = pd.Series(["" for _ in range(len(x1))])
    for i in range(0, len(pvalue_thresholds)):
        if i < len(pvalue_thresholds)-1:
            condition = (x1 <= pvalue_thresholds[i][0]) & (pvalue_thresholds[i+1][0] < x1)
            x_annot[condition] = pvalue_thresholds[i][1]
        else:
            condition = x1 < pvalue_thresholds[i][0]
            x_annot[condition] = pvalue_thresholds[i][1]

    return x_annot if not single_value else x_annot.iloc[0]


def simple_text(pval, pvalue_format, pvalue_thresholds, test_short_name=None):
    """
    Generates simple text for test name and pvalue
    :param pval: pvalue
    :param pvalue_format: format string for pvalue
    :param test_short_name: Short name of test to show
    :param pvalue_thresholds: String to display per pvalue range
    :return: simple annotation
    """
    # Sort thresholds
    thresholds = sorted(pvalue_thresholds, key=lambda x: x[0])

    # Test name if passed
    text = test_short_name and test_short_name + " " or ""

    for threshold in thresholds:
        if pval < threshold[0]:
            pval_text = "p ≤ {}".format(threshold[1])
            break
    else:
        pval_text = "p = {}".format(pvalue_format).format(pval)

    return text + pval_text


def add_stat_annotation(ax, plot='boxplot',
                        data=None, x=None, y=None, hue=None, units=None, order=None,
                        hue_order=None, box_pairs=None, width=0.8,
                        perform_stat_test=True,
                        pvalues=None, test_short_name=None,
                        test=None, text_format='star', pvalue_format_string=DEFAULT,
                        text_annot_custom=None,
                        loc='inside', show_test_name=True,
                        pvalue_thresholds=DEFAULT, stats_params=dict(),
                        comparisons_correction='bonferroni',
                        use_fixed_offset=False, line_offset_to_box=None,
                        line_offset=None, line_height=0.02, text_offset=1,
                        color='0.2', linewidth=1.5,
                        fontsize='medium', verbose=1):
    """
    Optionally computes statistical test between pairs of data series, and add statistical annotation on top
    of the boxes/bars. The same exact arguments `data`, `x`, `y`, `hue`, `order`, `width`,
    `hue_order` (and `units`) as in the seaborn boxplot/barplot function must be passed to this function.

    This function works in one of the two following modes:
    a) `perform_stat_test` is True: statistical test as given by argument `test` is performed.
    b) `perform_stat_test` is False: no statistical test is performed, list of custom p-values `pvalues` are
       used for each pair of boxes. The `test_short_name` argument is then used as the name of the
       custom statistical test.

    :param plot: type of the plot, one of 'boxplot' or 'barplot'.
    :param line_height: in axes fraction coordinates
    :param text_offset: in points
    :param box_pairs: can be of either form: For non-grouped boxplot: `[(cat1, cat2), (cat3, cat4)]`. For boxplot grouped by hue: `[((cat1, hue1), (cat2, hue2)), ((cat3, hue3), (cat4, hue4))]`
    :param pvalue_format_string: defaults to `"{.3e}"`
    :param pvalue_thresholds: list of lists, or tuples. Default is: For "star" text_format: `[[1e-4, "****"], [1e-3, "***"], [1e-2, "**"], [0.05, "*"], [1, "ns"]]`. For "simple" text_format : `[[1e-5, "1e-5"], [1e-4, "1e-4"], [1e-3, "0.001"], [1e-2, "0.01"]]`
    :param pvalues: list or array of p-values for each box pair comparison.
    :param comparisons_correction: Method for multiple comparisons correction. `bonferroni` or None.
    """

    def find_x_position_box(box_plotter, boxName):
        """
        boxName can be either a name "cat" or a tuple ("cat", "hue")
        """
        if box_plotter.plot_hues is None:
            cat = boxName
            hue_offset = 0
        else:
            cat = boxName[0]
            hue = boxName[1]
            hue_offset = box_plotter.hue_offsets[
                box_plotter.hue_names.index(hue)]

        group_pos = box_plotter.group_names.index(cat)
        box_pos = group_pos + hue_offset
        return box_pos

    def get_box_data(box_plotter, boxName):
        """
        boxName can be either a name "cat" or a tuple ("cat", "hue")

        Here we really have to duplicate seaborn code, because there is not
        direct access to the box_data in the BoxPlotter class.
        """
        cat = box_plotter.plot_hues is None and boxName or boxName[0]

        index = box_plotter.group_names.index(cat)
        group_data = box_plotter.plot_data[index]

        if box_plotter.plot_hues is None:
            # Draw a single box or a set of boxes
            # with a single level of grouping
            box_data = remove_na(group_data)
        else:
            hue_level = boxName[1]
            hue_mask = box_plotter.plot_hues[index] == hue_level
            box_data = remove_na(group_data[hue_mask])

        return box_data

    # Set default values if necessary
    if pvalue_format_string is DEFAULT:
        pvalue_format_string = '{:.3e}'
        simple_format_string = '{:.2f}'
    else:
        simple_format_string = pvalue_format_string

    if pvalue_thresholds is DEFAULT:
        if text_format == "star":
            pvalue_thresholds = [[1e-4, "****"], [1e-3, "***"],
                                 [1e-2, "**"], [0.05, "*"], [1, "ns"]]
        else:
            pvalue_thresholds = [[1e-5, "1e-5"], [1e-4, "1e-4"],
                                 [1e-3, "0.001"], [1e-2, "0.01"]]

    fig = plt.gcf()

    # Validate arguments
    if perform_stat_test:
        if test is None:
            raise ValueError("If `perform_stat_test` is True, `test` must be specified.")
        if pvalues is not None or test_short_name is not None:
            raise ValueError("If `perform_stat_test` is True, custom `pvalues` "
                             "or `test_short_name` must be `None`.")
        valid_list = ['t-test_ind', 't-test_welch', 't-test_paired',
                      'Mann-Whitney', 'Mann-Whitney-gt', 'Mann-Whitney-ls',
                      'Levene', 'Wilcoxon', 'Kruskal']
        if test not in valid_list:
            raise ValueError("test value should be one of the following: {}."
                             .format(', '.join(valid_list)))
    else:
        if pvalues is None:
            raise ValueError("If `perform_stat_test` is False, custom `pvalues` must be specified.")
        if test is not None:
            raise ValueError("If `perform_stat_test` is False, `test` must be None.")
        if len(pvalues) != len(box_pairs):
            raise ValueError("`pvalues` should be of the same length as `box_pairs`.")

    if text_annot_custom is not None and len(text_annot_custom) != len(box_pairs):
        raise ValueError("`text_annot_custom` should be of same length as `box_pairs`.")

    assert_is_in(
        loc, ['inside', 'outside'], label='argument `loc`'
    )
    assert_is_in(
        text_format,
        ['full', 'simple', 'star'],
        label='argument `text_format`'
    )
    assert_is_in(
        comparisons_correction,
        ['bonferroni', None],
        label='argument `comparisons_correction`'
    )

    if verbose >= 1 and text_format == 'star':
        print("p-value annotation legend:")
        pvalue_thresholds = pd.DataFrame(pvalue_thresholds).sort_values(by=0, ascending=False).values
        for i in range(0, len(pvalue_thresholds)):
            if i < len(pvalue_thresholds)-1:
                print('{}: {:.2e} < p <= {:.2e}'.format(pvalue_thresholds[i][1],
                                                        pvalue_thresholds[i+1][0],
                                                        pvalue_thresholds[i][0]))
            else:
                print('{}: p <= {:.2e}'.format(pvalue_thresholds[i][1], pvalue_thresholds[i][0]))
        print()

    ylim = ax.get_ylim()
    yrange = ylim[1] - ylim[0]

    if line_offset is None:
        if loc == 'inside':
            line_offset = 0.05
            if line_offset_to_box is None:
                line_offset_to_box = 0.06
        # 'outside', see valid_list
        else:
            line_offset = 0.03
            if line_offset_to_box is None:
                line_offset_to_box = line_offset
    else:
        if loc == 'inside':
            if line_offset_to_box is None:
                line_offset_to_box = 0.06
        elif loc == 'outside':
            line_offset_to_box = line_offset
    y_offset = line_offset*yrange
    y_offset_to_box = line_offset_to_box*yrange

    if plot == 'boxplot':
        # Create the same plotter object as seaborn's boxplot
        box_plotter = sns.categorical._BoxPlotter(
            x, y, hue, data, order, hue_order, orient=None, width=width, color=None,
            palette=None, saturation=.75, dodge=True, fliersize=5, linewidth=None)
    elif plot == 'barplot':
        # Create the same plotter object as seaborn's barplot
        box_plotter = sns.categorical._BarPlotter(
            x, y, hue, data, order, hue_order,
            estimator=np.mean, ci=95, n_boot=1000, units=None,
            orient=None, color=None, palette=None, saturation=.75,
            errcolor=".26", errwidth=None, capsize=None, dodge=True)

    # Build the list of box data structures with the x and ymax positions
    group_names = box_plotter.group_names
    hue_names = box_plotter.hue_names
    if box_plotter.plot_hues is None:
        box_names = group_names
        labels = box_names
    else:
        box_names = [(group_name, hue_name) for group_name in group_names for hue_name in hue_names]
        labels = ['{}_{}'.format(group_name, hue_name) for (group_name, hue_name) in box_names]

    box_structs = [{'box':box_names[i],
                    'label':labels[i],
                    'x':find_x_position_box(box_plotter, box_names[i]),
                    'box_data':get_box_data(box_plotter, box_names[i]),
                    'ymax':np.amax(get_box_data(box_plotter, box_names[i])) if
                           len(get_box_data(box_plotter, box_names[i])) > 0 else np.nan}
                   for i in range(len(box_names))]
    # Sort the box data structures by position along the x axis
    box_structs = sorted(box_structs, key=lambda x: x['x'])
    # Add the index position in the list of boxes along the x axis
    box_structs = [dict(box_struct, xi=i) for i, box_struct in enumerate(box_structs)]
    # Same data structure list with access key by box name
    box_structs_dic = {box_struct['box']:box_struct for box_struct in box_structs}

    # Build the list of box data structure pairs
    box_struct_pairs = []
    for i_box_pair, (box1, box2) in enumerate(box_pairs):
        valid = box1 in box_names and box2 in box_names
        if not valid:
            raise ValueError("box_pairs contains an invalid box pair.")
            pass
        # i_box_pair will keep track of the original order of the box pairs.
        box_struct1 = dict(box_structs_dic[box1], i_box_pair=i_box_pair)
        box_struct2 = dict(box_structs_dic[box2], i_box_pair=i_box_pair)
        if box_struct1['x'] <= box_struct2['x']:
            pair = (box_struct1, box_struct2)
        else:
            pair = (box_struct2, box_struct1)
        box_struct_pairs.append(pair)

    # Draw first the annotations with the shortest between-boxes distance, in order to reduce
    # overlapping between annotations.
    box_struct_pairs = sorted(box_struct_pairs, key=lambda x: abs(x[1]['x'] - x[0]['x']))

    # Build array that contains the x and y_max position of the highest annotation or box data at
    # a given x position, and also keeps track of the number of stacked annotations.
    # This array will be updated when a new annotation is drawn.
    y_stack_arr = np.array([[box_struct['x'] for box_struct in box_structs],
                            [box_struct['ymax'] for box_struct in box_structs],
                            [0 for i in range(len(box_structs))]])
    if loc == 'outside':
        y_stack_arr[1, :] = ylim[1]
    ann_list = []
    test_result_list = []
    ymaxs = []
    y_stack = []

    for box_struct1, box_struct2 in box_struct_pairs:

        box1 = box_struct1['box']
        box2 = box_struct2['box']
        label1 = box_struct1['label']
        label2 = box_struct2['label']
        box_data1 = box_struct1['box_data']
        box_data2 = box_struct2['box_data']
        x1 = box_struct1['x']
        x2 = box_struct2['x']
        xi1 = box_struct1['xi']
        xi2 = box_struct2['xi']
        ymax1 = box_struct1['ymax']
        ymax2 = box_struct2['ymax']
        i_box_pair = box_struct1['i_box_pair']

        # Find y maximum for all the y_stacks *in between* the box1 and the box2
        i_ymax_in_range_x1_x2 = xi1 + np.nanargmax(y_stack_arr[1, np.where((x1 <= y_stack_arr[0, :]) &
                                                                           (y_stack_arr[0, :] <= x2))])
        ymax_in_range_x1_x2 = y_stack_arr[1, i_ymax_in_range_x1_x2]

        if perform_stat_test:
            result = stat_test(
                box_data1,
                box_data2,
                test,
                comparisons_correction,
                len(box_struct_pairs),
                **stats_params
            )
        else:
            test_short_name = test_short_name if test_short_name is not None else ''
            result = StatResult(
                'Custom statistical test',
                test_short_name,
                None,
                None,
                pvalues[i_box_pair]
            )

        result.box1 = box1
        result.box2 = box2
        test_result_list.append(result)

        if verbose >= 1:
            print("{} v.s. {}: {}".format(label1, label2, result.formatted_output))

        if text_annot_custom is not None:
            text = text_annot_custom[i_box_pair]
        else:
            if text_format == 'full':
                text = "{} p = {}".format('{}', pvalue_format_string).format(result.test_short_name, result.pval)
            elif text_format is None:
                text = None
            elif text_format is 'star':
                text = pval_annotation_text(result.pval, pvalue_thresholds)
            elif text_format is 'simple':
                test_short_name = show_test_name and test_short_name or ""
                text = simple_text(result.pval, simple_format_string, pvalue_thresholds, test_short_name)

        yref = ymax_in_range_x1_x2
        yref2 = yref

        # Choose the best offset depending on wether there is an annotation below
        # at the x position in the range [x1, x2] where the stack is the highest
        if y_stack_arr[2, i_ymax_in_range_x1_x2] == 0:
            # there is only a box below
            offset = y_offset_to_box
        else:
            # there is an annotation below
            offset = y_offset
        y = yref2 + offset
        h = line_height*yrange
        line_x, line_y = [x1, x1, x2, x2], [y, y + h, y + h, y]
        if loc == 'inside':
            ax.plot(line_x, line_y, lw=linewidth, c=color)
        elif loc == 'outside':
            line = lines.Line2D(line_x, line_y, lw=linewidth, c=color, transform=ax.transData)
            line.set_clip_on(False)
            ax.add_line(line)

        # why should we change here the ylim if at the very end we set it to the correct range????
        # ax.set_ylim((ylim[0], 1.1*(y + h)))

        if text is not None:
            ann = ax.annotate(
                text, xy=(np.mean([x1, x2]), y + h),
                xytext=(0, text_offset), textcoords='offset points',
                xycoords='data', ha='center', va='bottom',
                fontsize=fontsize, clip_on=False, annotation_clip=False)
            ann_list.append(ann)

            plt.draw()
            y_top_annot = None
            got_mpl_error = False
            if not use_fixed_offset:
                try:
                    bbox = ann.get_window_extent()
                    bbox_data = bbox.transformed(ax.transData.inverted())
                    y_top_annot = bbox_data.ymax
                except RuntimeError:
                    got_mpl_error = True

            if use_fixed_offset or got_mpl_error:
                if verbose >= 1:
                    print("Warning: cannot get the text bounding box. Falling back to a fixed"
                          " y offset. Layout may be not optimal.")
                # We will apply a fixed offset in points,
                # based on the font size of the annotation.
                fontsize_points = FontProperties(size='medium').get_size_in_points()
                offset_trans = mtransforms.offset_copy(
                    ax.transData, fig=fig, x=0,
                    y=1.0*fontsize_points + text_offset, units='points')
                y_top_display = offset_trans.transform((0, y + h))
                y_top_annot = ax.transData.inverted().transform(y_top_display)[1]
        else:
            y_top_annot = y + h

        y_stack.append(y_top_annot)    # remark: y_stack is not really necessary if we have the stack_array
        ymaxs.append(max(y_stack))
        # Fill the highest y position of the annotation into the y_stack array
        # for all positions in the range x1 to x2
        y_stack_arr[1, (x1 <= y_stack_arr[0, :]) & (y_stack_arr[0, :] <= x2)] = y_top_annot
        # Increment the counter of annotations in the y_stack array
        y_stack_arr[2, xi1:xi2 + 1] = y_stack_arr[2, xi1:xi2 + 1] + 1

    y_stack_max = max(ymaxs)
    if loc == 'inside':
        ax.set_ylim((ylim[0], max(1.03*y_stack_max, ylim[1])))
    elif loc == 'outside':
        ax.set_ylim((ylim[0], ylim[1]))

    return ax, test_result_list



import torch
import numpy as np
import torch
import torch.nn.functional as F
def calc_contrast_loss(features,assignment_orig,device):

    # Use a batch size of 1000
    nbatch = 10000
    sel_vals = torch.randint(0,features.shape[0],(nbatch,1))

    # Calculate feature similarities
    features = F.normalize(features[sel_vals].squeeze(), dim=1)
    features = torch.matmul(features, features.T)

    # Find top-1
    features[torch.eye(features.shape[0], dtype=torch.bool).to(device)] = 0
    top_elem = features.topk(1,1)[1]
    #low_elem = features.topk(1,1,largest=False)[1]

    # Calculate assignment similarities  
    assignment = assignment_orig[sel_vals].squeeze().clone()
    bound = assignment.max(1,True)[0]
    assignment[assignment != bound] = 0  # set the values i want to keep to 1
    assignment[assignment == bound] = 1            

    # Contrast 
    temp = 0.07
    positives = torch.exp((assignment*assignment[top_elem.squeeze(),:]).sum(1)/temp)
    #negatives = torch.exp((assignment*assignment[torch.randint(0,assignment.shape[0],(assignment.shape[0],1)).squeeze(),:]).sum(1)/temp)

    # # Accuracy
    # pos = torch.eq(assignment.argmax(1),assignment[top_elem.squeeze(),:].argmax(1)).sum()/assignment.shape[0]
    # neg = torch.eq(assignment.argmax(1),assignment[low_elem.squeeze(),:].argmax(1)).sum()/assignment.shape[0]    
    
    return -torch.log(positives).mean()*temp

def calc_patch_ent(s,device):    
    return ((-s/s.sum(dim=-1,keepdim=True) * torch.log(s/s.sum(dim=-1,keepdim=True) + 1e-15)).sum(dim=-1)/torch.log(torch.tensor(s.shape[1],dtype=torch.float32,device=device))).mean()   

def patch_contrast_entropy(x, s, num_nodes, device):
    
    # Calculate contrast acc        
    contrast_loss = calc_contrast_loss(torch.cat([x[i,:num_nodes[i],:] for i in range(s.shape[0])]), torch.cat([s[i,:num_nodes[i],:] for i in range(s.shape[0])]),device)                     
    
    # Calculate Patch Entropy
    PatchEntropy_loss = calc_patch_ent(torch.cat([s[i,:num_nodes[i],:] for i in range(s.shape[0])]),device)                     

    return contrast_loss, PatchEntropy_loss
    
def Patient_entropy(args,S,device):
    # Keep High Entropy in the network
    if args['Max_Pat_Entropy']:                                          
        NN_loss1 = ((S[0]/S[0].sum(-1,keepdim=True)*torch.log(S[0]/S[0].sum(-1,keepdim=True)+ 1e-15)).sum(dim=1)/torch.log(torch.tensor(S[0].shape[1],dtype=torch.float32,device=device))).mean()
        NN_loss2 = ((S[1]/S[1].sum(-1,keepdim=True)*torch.log(S[1]/S[1].sum(-1,keepdim=True)+ 1e-15)).sum(dim=1)/torch.log(torch.tensor(S[1].shape[1],dtype=torch.float32,device=device))).mean()
        NN_loss3 = ((S[2]/S[2].sum(-1,keepdim=True)*torch.log(S[2]/S[2].sum(-1,keepdim=True)+ 1e-15)).sum(dim=1)/torch.log(torch.tensor(S[2].shape[1],dtype=torch.float32,device=device))).mean()
        print('Entropy Phenotypes: ',NN_loss1.item(),' Entropy TissueCommunities: ',NN_loss2.item(), 'Entropy TissueCommunitiesInter: ',NN_loss3.item(),'Min_Clust: ',S[0].sum(-2).min().item())
        return NN_loss1, NN_loss2, NN_loss3
    else:
        return 0, 0, 0

def NearestNeighborClassification(self):
    NN_loss, _, _, _ = utilz.nearestNeighbor_loss(self.training, args, self.sPatient[-1], data.y.long(), self.NNCPosition,index,Indices,labels,device,self)                                
    NN_loss = -(-torch.cat((self.S[0],self.S[1],self.S[2]),dim=1) * torch.log(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1) + 1e-15)).sum(dim=-1).mean()                

def f_test_loss(features,labels,device,lin1,lin2,BN,args):
    
    if args['F-test']:
        # Unrestricted Model
        UM = F.relu(lin1(features)) # torch.cat(self.X,dim=-1)            
        UM = lin2(UM)         
        # Model Loss of Unrestricted Model
        unrestricted_loss = F.cross_entropy(UM, labels[:,0].long())     
        pred = UM.max(1)[1].detach().cpu().numpy()
        print('train_acc:',np.equal(pred,labels[:,0].cpu().numpy()).mean())

        # # Sum of Squares Residuals of unrestricted model
        # unrestricted_RSS = torch.pow((torch.matmul(features,regression_model.weight.t())+regression_model.bias).squeeze()-labels,2).sum()    
        # # Leave-one-out f-test model
        restricted_loss = torch.zeros(features.shape[1],device=device)
        
        # For copies
        # W_f = regression_model.weight.clone()
        indices = torch.arange(features.shape[1],device=device)
        
        # Restrict model per feature.
        for f in range(features.shape[1]):
            # Put Weight of f_index to zero        
            # regression_model.weight[0,f] = 0

            # Obtain Sum of Squares of restricted model        
            Restricted_Info = torch.zeros(features.shape,device=device)        
            Restricted_Info[:,indices!=f] = features[:,indices!=f]
            restricted_loss[f] = F.cross_entropy(lin2(BN(F.relu(lin1(Restricted_Info)))), labels[:,0].long()) if Restricted_Info.shape[0]>1 else F.cross_entropy(lin2(F.relu(lin1(Restricted_Info))), labels[:,0].long())        
            # torch.cat(self.X,dim=-1)                 
            
            
            # restricted_RSS = torch.pow((torch.matmul(features[:,x[x!=f]],regression_model.weight[:,x[x!=f]].t())+regression_model.bias).squeeze()-labels,2).sum()

            # F-test
            # loss[f] = (restricted_RSS-unrestricted_RSS)/(features.shape[0]-features.shape[1])

        # loss = F.relu(loss)
        # Sparsify the unrestricted models
        restricted_loss = (restricted_loss/unrestricted_loss)  
        restricted_loss = restricted_loss/restricted_loss.sum()  
        restricted_loss = -(restricted_loss * torch.log(restricted_loss + 1e-15)).mean()
        # restricted_loss = -restricted_loss.max()
        # restricted_loss = -(torch.sqrt(restricted_loss.shape[0]*torch.ones(1,device=device))-(restricted_loss.sum()/(torch.pow(restricted_loss,2).sum()+1e-16)))/(torch.sqrt(restricted_loss.shape[0]*torch.ones(1,device=device))-1)
        print('sparsify the f-test',restricted_loss, 'loss of original Model',unrestricted_loss)
        f_test_loss = unrestricted_loss+restricted_loss
    else:
        f_test_loss = 0

    return f_test_loss

def SupConLoss(features,labels,device,temperature=0.07):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    it degenerates to SimCLR unsupervised loss:
    https://arxiv.org/pdf/2002.05709.pdf
    Args:
        features: hidden vector of shape [bsz, n_views, ...].
        labels: ground truth of shape [bsz].
        mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
            has the same class as sample i. Can be asymmetric.
    Returns:
        A loss scalar.
    """
    # device = (torch.device('cuda')
    #           if features.is_cuda
    #           else torch.device('cpu'))
    
    loss = torch.zeros(1,device=device)
    features = features.unsqueeze(dim=1)
    Contrast_acc = 0
    n_iter = 0

    # Calculate loss only if there are repeated values in loss
    a, rep = np.unique(labels.cpu().numpy(),return_counts=True)
    if any(rep>1):
        # Iterate Over each patient i.        
        for i in range(features.shape[0]): 
            
            # Reinitialize patient_i_loss
            patient_i_loss = torch.zeros(1,device=device)

            # Iterate Over each patient j.
            for j in range(features.shape[0]): 
                
                # Select patient if they have same labels.
                if (i!=j) and (labels[i]==labels[j]):                 
                    # Apply supContrast Loss
                    patient_i_loss += torch.log(
                        torch.exp(F.cosine_similarity(features[i],features[j])/temperature)/
                        torch.exp(F.cosine_similarity(features[i],features[labels[i]!=labels].squeeze(dim=1))/temperature).mean())
                    Contrast_acc += F.cosine_similarity(features[i],features[j])-F.cosine_similarity(features[i],features[labels[i]!=labels].squeeze(dim=1)).mean()
                    n_iter += 1 

            # Normalize patient_i_loss
            loss += (-torch.ones(1,device=device)/((labels[i]==labels).sum()-1+1e-16))*patient_i_loss
        
        loss /= features.shape[0]
                    
    return loss, Contrast_acc/(n_iter+1e-15)

def SupConLoss_Total(args,S,labels,data,device):
    if args['orthoColor']:
        # supconloss = torch.relu(self.SUPCONlin(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1)))
        # supconloss = utilz.SupConLoss(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1),data.y,device,temperature=0.07)
        if args['F-test']:
            supconloss = SupConLoss(S[0][:,:max(labels)+1],data.y,device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[0][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
            supconloss += SupConLoss(S[1][:,:max(labels)+1],data.y,device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[1][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
            supconloss += SupConLoss(S[2][:,:max(labels)+1],data.y,device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[2][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
            # f_test_loss = -self.S[0][:,max(labels)+1:].mean()
            # f_test_loss += -self.S[1][:,max(labels)+1:].mean()
            # f_test_loss += -self.S[2][:,max(labels)+1:].mean()
            # f_test_loss = f_test_loss*0.001
        else:
            supconloss1, sup_acc1  = SupConLoss(S[0],data.y[:,0],device,temperature=0.07)
            supconloss2, sup_acc2 = SupConLoss(S[1],data.y[:,0],device,temperature=0.07)
            supconloss3, sup_acc3 = SupConLoss(S[2],data.y[:,0],device,temperature=0.07)
            if np.argmin([sup_acc1,sup_acc2,sup_acc3])==0:
                supconloss=supconloss1
            elif np.argmin([sup_acc1,sup_acc2,sup_acc3])==1:
                supconloss=supconloss2
            else:
                supconloss=supconloss3
            print('Supervised_Contrast_accuracy: ',(sup_acc1).item(),+sup_acc2.item(), sup_acc3.item())
    else:
        supconloss=0

    return supconloss

def unsup_ColorLoss(projection,labels, features,device,temperature=0.07):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    it degenerates to SimCLR unsupervised loss:
    https://arxiv.org/pdf/2002.05709.pdf
    Args:
        features: hidden vector of shape [bsz, n_views, ...].
        labels: ground truth of shape [bsz].
        mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
            has the same class as sample i. Can be asymmetric.
    Returns:
        A loss scalar.
    """
    # device = (torch.device('cuda')
    #           if features.is_cuda
    #           else torch.device('cpu'))
    
    loss = torch.zeros(1,device=device)
    projection = projection/projection.max()
    projection = projection.unsqueeze(dim=1)
    features = features/features.max()
    features = features.unsqueeze(dim=1)

    unsup_Contrast_acc = 0 
    n_iter = 0

    # Calculate loss only if there are more than One Image.
    if labels.shape[0]>2:
        # Iterate Over each patient i.        
        for i in list(range(0,projection.shape[0],1)): 
            
            # Reinitialize patient_i_loss
            patient_i_loss = torch.zeros(1,device=device)

            # Iterate Over each patient j.
            for j in range(projection.shape[0]): 
                
                # Select patient if they have same labels.
                if (i!=j):                 
                    # Apply supContrast Loss
                    patient_i_loss += torch.exp(F.cosine_similarity(projection[i],projection[j]))
                    # unsup_Contrast_acc += F.cosine_similarity(features[i],features[i+1])-F.cosine_similarity(features[i],features[j])
                    n_iter += 1 

            # Normalize patient_i_loss
            loss += patient_i_loss
        
        loss /= projection.shape[0]*(projection.shape[0]-1)
                    
    return loss, unsup_Contrast_acc/(n_iter+1e-16)

def UnsupConLoss(projection,labels, features,device,temperature=0.07):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    it degenerates to SimCLR unsupervised loss:
    https://arxiv.org/pdf/2002.05709.pdf
    Args:
        features: hidden vector of shape [bsz, n_views, ...].
        labels: ground truth of shape [bsz].
        mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
            has the same class as sample i. Can be asymmetric.
    Returns:
        A loss scalar.
    """
    # device = (torch.device('cuda')
    #           if features.is_cuda
    #           else torch.device('cpu'))
    
    loss = torch.zeros(1,device=device)

    projection = projection.unsqueeze(dim=1)
    features = features.unsqueeze(dim=1)

    unsup_Contrast_acc = 0 
    n_iter = 0

    # Calculate loss only if there are more than One Image.
    if labels.shape[0]>2:
        # Iterate Over each patient i.        
        for i in list(range(0,projection.shape[0],2)): 
            
            # Reinitialize patient_i_loss
            patient_i_loss = torch.zeros(1,device=device)

            # Iterate Over each patient j.
            for j in range(projection.shape[0]): 
                
                # Select patient if they have same labels.
                if (i!=j) and ((i+1)!=j):                 
                    # Apply supContrast Loss
                    patient_i_loss += -torch.log(
                        torch.exp(F.cosine_similarity(projection[i],projection[i+1])/temperature)/
                        torch.exp(F.cosine_similarity(projection[i],projection[j])/temperature).mean())
                    unsup_Contrast_acc += F.cosine_similarity(features[i],features[i+1])-F.cosine_similarity(features[i],features[j])
                    n_iter += 1 

            # Normalize patient_i_loss
            loss += patient_i_loss
        
        loss /= projection.shape[0]*(projection.shape[0]-2)
                    
    return loss, unsup_Contrast_acc/(n_iter+1e-16)

def UnsupConLoss_Total(args, S, data, device, self):
    if args['UnsupContrast'] and self.training:
        # unsupconloss1, unsup_acc1 = utilz.UnsupConLoss(self.lin1_unsupB(F.relu(self.lin1_unsupA(self.S[0]))),data.y,self.S[0],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[0][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        # unsupconloss2, unsup_acc2 = utilz.UnsupConLoss(self.lin2_unsupB(F.relu(self.lin2_unsupA(self.S[1]))),data.y, self.S[1],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[1][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        # unsupconloss3, unsup_acc3 = utilz.UnsupConLoss(self.lin3_unsupB(F.relu(self.lin3_unsupA(self.S[2]))),data.y,self.S[2],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[2][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        unsupconloss1, unsup_acc1 = UnsupConLoss(S[0],data.y,S[0],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[0][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        unsupconloss2, unsup_acc2 = UnsupConLoss(S[1],data.y,S[1],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[1][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        unsupconloss3, unsup_acc3 = UnsupConLoss(S[2],data.y,S[2],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[2][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                    
        # if np.argmin([unsup_acc1,unsup_acc2,unsup_acc3])==0:
        #     unsupconloss=unsupconloss1
        # elif np.argmin([unsup_acc1,unsup_acc2,unsup_acc3])==1:
        #     unsupconloss=unsupconloss2
        # else:
        #     unsupconloss=unsupconloss2
        unsup_acc = (unsup_acc1+unsup_acc2+unsup_acc3)/3
        # else:
        #     unsupconloss=unsupconloss3
        # print('Unsupervised_accuracy: ',(unsup_acc1).item(),+unsup_acc2.item(), unsup_acc3.item())
        print('Unsupervised_accuracy: ',unsup_acc)
        return unsupconloss1, unsup_acc1, unsupconloss2, unsup_acc2, unsupconloss3, unsup_acc3 
    else: 
        return 0,0,0,0,0,0

def Lasso_Feat_Selection(layer1,layer2):
    return 0.1*layer1.weight.abs().sum() + 0.1*layer2.weight.abs().sum()


    




import logging
import os
import sys

import torch
import torch.nn.functional as F
from torch.cuda.amp import GradScaler, autocast
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from NaroNet.Patch_Contrastive_Learning.utils import save_config_file, accuracy, save_checkpoint
import pandas as pd
import numpy as np
from PIL import Image
from sklearn.cluster import AgglomerativeClustering
import gc 
gc.set_threshold(0,0,0)

torch.manual_seed(0)

def visualize_using_clustering(name,features,image_size,patch_size,image):

    # Create patch Image
    Patch_im = np.zeros(image_size)                                
    division_rows = np.floor(Patch_im.shape[0]/patch_size)
    division_cols = np.floor(Patch_im.shape[1]/patch_size)
    lins = np.repeat(list(range(int(division_cols))), patch_size)
    lins = np.repeat(np.expand_dims(lins,axis=1), patch_size, axis=1)
    for row_indx, row in enumerate(range(int(division_rows))):
        Patch_im[row_indx*patch_size:(row_indx+1)*patch_size,:int(division_cols*patch_size)] = np.transpose(lins+int(row_indx*division_cols))    
    Patch_im = Patch_im.astype(int)

    # Cluster into distinct phenotypes
    clustering = AgglomerativeClustering(n_clusters=3).fit(features)
    Cluster_image = clustering.labels_[Patch_im]

    # Save patch Image
    imtosave = Image.fromarray(np.uint16(Patch_im))
    imtosave.save(name+'_PatchIm.png')                

    # Save cluster image
    imtosave = Image.fromarray(np.uint8(Cluster_image))
    imtosave.save(name+'_cluster.png') 

    # Save image    
    imtosave = Image.fromarray(np.uint16((image.squeeze()[:,:,6].numpy()/image.squeeze()[:,:,6].numpy().max())*65550))
    imtosave.save(name+'_image.png')                


class SimCLR(object):

    def __init__(self, *args, **kwargs):
        self.args = kwargs['args']
        self.model = kwargs['model'].to(self.args['PCL_GPU_INDEX'])
        self.optimizer = kwargs['optimizer']
        self.scheduler = kwargs['scheduler']            
        self.model_dir = self.args['path']+'Patch_Contrastive_Learning/Model/'        
        self.writer = SummaryWriter(log_dir=self.model_dir)    
        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir)  
        logging.basicConfig(filename=os.path.join(self.model_dir, 'training.log'), level=logging.DEBUG)
        self.criterion = torch.nn.CrossEntropyLoss().to(self.args['PCL_GPU_INDEX'])

    def info_nce_loss(self, features):

        labels = torch.cat([torch.arange(features.shape[0]/2) for i in range(2)], dim=0)
        labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
        labels = labels.to(self.args['PCL_GPU_INDEX'])

        features = F.normalize(features, dim=1)

        similarity_matrix = torch.matmul(features, features.T)

        # discard the main diagonal from both: labels and similarities matrix
        mask = torch.eye(labels.shape[0], dtype=torch.bool).to(self.args['PCL_GPU_INDEX'])
        labels = labels[~mask].view(labels.shape[0], -1)
        similarity_matrix = similarity_matrix[~mask].view(similarity_matrix.shape[0], -1)

        # select and combine multiple positives
        positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)

        # select only the negatives the negatives
        negatives = similarity_matrix[~labels.bool()].view(similarity_matrix.shape[0], -1)

        logits = torch.cat([positives, negatives], dim=1)
        labels = torch.zeros(logits.shape[0], dtype=torch.long).to(self.args['PCL_GPU_INDEX'])

        logits = logits / self.args['PCL_Temperature']
        return logits, labels

    def train(self, train_loader):

        scaler = GradScaler(enabled=True)

        # save config file
        save_config_file(self.writer.log_dir, self.args)

        n_iter = 0
        logging.info(f"Start SimCLR training for {self.args['PCL_Epochs']} epochs.")        

        for epoch_counter in tqdm(range(self.args['PCL_Epochs'])):
            for images in train_loader:
                images[0] = torch.cat(torch.tensor_split(images[0], images[0].shape[0]),1).squeeze()
                images[1] = torch.cat(torch.tensor_split(images[1], images[1].shape[0]),1).squeeze()

                images = torch.cat((images[0],images[1]),dim=0)

                images = images.to(torch.float32)

                images = images.to(self.args['PCL_GPU_INDEX'])

                images = torch.moveaxis(images, 3, 1)

                with autocast(enabled=True):
                    features = self.model(images)
                    logits, labels = self.info_nce_loss(features)
                    loss = self.criterion(logits, labels)
                
                self.optimizer.zero_grad()

                scaler.scale(loss).backward()

                scaler.step(self.optimizer)
                scaler.update()

                # Log info
                top1, top5 = accuracy(logits, labels, topk=(1, 5))
                self.writer.add_scalar('loss', loss, global_step=n_iter)
                self.writer.add_scalar('acc/top1', top1[0], global_step=n_iter)
                self.writer.add_scalar('acc/top5', top5[0], global_step=n_iter)
                self.writer.add_scalar('learning_rate', self.scheduler.get_lr()[0], global_step=n_iter)

                n_iter += 1                
                
                del images, features
                gc.collect(0)
                gc.collect(1)
                gc.collect(2)
                logging.debug(f"Epoch: {epoch_counter}\tLoss: {loss}\tTop1 accuracy: {top1[0]}\tTop5 accuracy: {top5[0]}")

            # warmup for the first 10 epochs
            if epoch_counter >= 6*(self.args['PCL_Epochs']/10):
                self.scheduler.step()
            
            # save model checkpoints
            if epoch_counter % int(self.args['PCL_Epochs']/10) == 0:
                checkpoint_name = 'checkpoint_{:04d}.pth.tar'.format(epoch_counter)
                torch.save(self.model.state_dict(), os.path.join(self.model_dir, checkpoint_name))        
                logging.info(f"Model checkpoint and metadata has been saved at {self.model_dir}.")

        logging.info("Training has finished.")
        

    def infer(self,infer_loader, out_dir):
        
        # Open last saved model checkpoint          
        text_files = [f for f in os.listdir(self.model_dir) if f.endswith('.tar')]
        text_files.sort()
        self.model.load_state_dict(torch.load(os.path.join(self.model_dir, text_files[-1])))
        del self.model.backbone.fc._modules['4']
        del self.model.backbone.fc._modules['3']
        self.model.eval()       

        self.folder = out_dir 

        # Order Images 
        infer_loader.dataset.image_list.sort()

        # Extract patches and save it into a graph
        for patches, patch_pos, filename in tqdm(infer_loader):                                    
            
            # Locate model to device.
            self.model = self.model.to(self.args['PCL_GPU_INDEX'])
            
            # Minibatch to extract features
            step = 400
            features = []
            for p in range(0,len(patches),step):                
                features.append(self.model(torch.moveaxis(torch.squeeze(torch.tensor(np.stack(patches[p:p+step],0),device=self.args['PCL_GPU_INDEX'])),3,1).to(torch.float32)).to('cpu').detach().numpy())
                gc.collect(0)
                gc.collect(1)      
                gc.collect(2)      
            features = np.concatenate(features,axis=0)

            # Save features
            features = np.concatenate((np.squeeze(patch_pos),features),axis=1)    

            # Join several images to one numpy structure
            patient_to_image = [i for i in os.listdir(self.args['path']+'Raw_Data/Experiment_Information/') if 'Patient_to_Image.xlsx'==i]
            if len(patient_to_image)>0:
                
                # Find subject name and subject file
                patient_to_image_excel = pd.read_excel(self.args['path']+'Raw_Data/Experiment_Information/'+patient_to_image[0])
                index = list(patient_to_image_excel['Image_Name']).index(filename[0])
                NPYname = self.folder+patient_to_image_excel['Subject_Name'][index]+'.npy'
                
                # if already exists...                 
                if os.path.isfile(NPYname):
                    
                    # Load previous concatenate with actual data and save it.
                    prevFeats = np.load(NPYname)
                    features[:,:2] = features[:,:2] + prevFeats[:,:2].max() + self.args['PCL_Patch_Size']*10 
                    np.save(NPYname,np.concatenate((prevFeats,features),axis=0))      
                    del prevFeats           

                else:

                    # Create new file
                    np.save(NPYname,features)    
                
            # Save one image is one subject
            else:
                np.save(self.folder+filename[0],features)    
                # tifffile.imsave(self.folder+filename[0]+'##.tiff',np.stack([np.reshape(features[:,i],(289,240)) for i in range(features.shape[1])]))


            # Eliminate data from garbage collector
            del patches, patch_pos, filename, features 
            gc.collect(0)
            gc.collect(1)      
            gc.collect(2)      
            
            
            # visualize_using_clustering(folder+filename[0],features,[1200,1200],patches.shape[2],image)
            
            
            

        return 



import numpy as np
import os.path as osp
import copy
import os
import matplotlib.pyplot as plt
import cv2
from skimage.morphology import opening, disk
from skimage.measure import perimeter
from matplotlib import cm
from scipy import stats
import seaborn as sns
from NaroNet.utils.parallel_process import parallel_process
from tifffile.tifffile import imwrite
import pandas as pd
import gc
import itertools
from skimage import filters
import time 

def load_cell_types_assignments(dataset, cell_type_idx, subject_info,subgraph_idx,n_cell_types, prev_cell_type_assignment):
    """
    Obtain matrix that assigns patches to cell types (phenotypes and neighborhoods)
    dataset: (object)
    cell_type_idx: (int) cell_type_idx==0 is phenotype, cell_type_idx==1 is neihborhood, cell_type_idx==2 is neihborhood interaction
    subject_info: (list of str and int)
    subgraph_idx: (int) specifying the index of the subgraph.
    n_cell_types: (int) number of cell types (phenotypes or neighborhoods)
    prev_cell_type_assignment: (array of int) specifying assignments
    """

    # If phenotype or neighborhood load matrix assignment.
    if cell_type_idx<2:
        cell_type_assignment = np.load(osp.join(dataset.processed_dir_cell_types,
                    'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], subgraph_idx, n_cell_types)))                        

    # if neighborhood interaction the matrix assigns neighbors to neighborhood interactions.
    else:        
        # Load neighborhood interactions
        secondorder_assignment = np.load(osp.join(dataset.processed_dir_cell_types,
                                'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(subject_info[1], n_cell_types)))                    
        
        # Obtain assignment of patches to neighborhood interactions
        cell_type_assignment = np.matmul(prev_cell_type_assignment,secondorder_assignment)
    prev_cell_type_assignment = copy.deepcopy(cell_type_assignment)
    
    return prev_cell_type_assignment, cell_type_assignment

def load_patch_image(dataset,imList):
    '''
    Load patch image
    # Create label map. 
    labels = []
    last_patch = 0
    for im_i in imList:
        num_patches = int(im_i.shape[0]/dataset.patch_size)*int(im_i.shape[1]/dataset.patch_size)
        labels.append(np.reshape(np.array(range(last_patch,last_patch+num_patches)),(int(im_i.shape[0]/dataset.patch_size),int(im_i.shape[1]/dataset.patch_size))))
        last_patch = last_patch+num_patches
    labels = np.concatenate(labels,1)
    '''             

    labels = []
    last_patch = 0
    for im_i in imList:
        num_patches = int(im_i.shape[0]/dataset.patch_size)*int(im_i.shape[1]/dataset.patch_size)
        labels.append(np.transpose(np.reshape(np.array(range(last_patch,last_patch+num_patches)),(int(im_i.shape[0]/dataset.patch_size),int(im_i.shape[1]/dataset.patch_size)))))
        last_patch = last_patch+num_patches

    return np.concatenate(labels,1), labels

def select_patches_from_cohort_(clusters,dataset,subject_info,count):

    # Crops,Confidence,and Phenotype Vector.
    CropConfPheno = []
    CropConfTissueComm = []
    TissueArea_Conf = []
    for c in range(clusters[0]):
        CropConfPheno.append([])
            
    for c in range(clusters[1]):
        CropConfTissueComm.append([])  
    
    for c in range(clusters[2]):
        TissueArea_Conf.append([]) 

    # Crops,Confidence,and Phenotype Vector.    
    prev_cell_type_assignment=[]
        
    # Apply mask to patch
    for subgraph_idx in range(dataset.findLastIndex(subject_info[1])+1):                
        
        # Open Raw Image.        
        im, imList = dataset.open_Raw_Image(subject_info,1)                

        for cell_type_idx, n_cell_types in enumerate(clusters):    
            # load cell_types_assignments
            prev_cell_type_assignment, cell_type_assignment = load_cell_types_assignments(
                                                                    dataset,cell_type_idx,subject_info,subgraph_idx,
                                                                    n_cell_types,prev_cell_type_assignment)                               
            # Open Single-Cell Contrastive Learning Information
            PCL_reprsntions = np.load(dataset.raw_dir+'/{}.npy'.format(subject_info[0]))                     
            # Select Patches of a specific Phenotype and save its confidence
            cell_type_top1 = np.apply_along_axis(np.argmax, axis=1, arr=cell_type_assignment) 
            
            # Phenotypes and neighborhoods, or areas
            if cell_type_idx<2:             
                # Select patches with most confidence
                CropConfPheno,CropConfTissueComm = topk_confident_patches(dataset,n_cell_types,cell_type_top1,
                                                                            cell_type_assignment,cell_type_idx,CropConfPheno,
                                                                            CropConfTissueComm,im,imList,PCL_reprsntions,count)  
            else:
                # Save subject index and max cell type
                for c_type in range(n_cell_types):
                    TissueArea_Conf[c_type].append(cell_type_assignment[:,c_type])

    del im, imList, PCL_reprsntions, cell_type_top1
    gc.collect(0)
    gc.collect(1)
    gc.collect(2)
    time.sleep(3)

    return CropConfPheno,CropConfTissueComm,TissueArea_Conf


def select_patches_from_cohort(dataset,IndexAndClass,clusters):
    '''
    '''
     
    # Prepare parallel process
    dict_subjects = []
    for count, subject_info in enumerate(IndexAndClass):
        if subject_info[2][0]!='None':
            dict_subjects.append({'clusters':clusters,'dataset':dataset,'subject_info':subject_info,'count':count})
    
    # select_patches_from_cohort
    result_images = parallel_process(dict_subjects,select_patches_from_cohort_,use_kwargs=True,front_num=1,desc='BioInsights: Get relevant examples of cell types') 

    # Crops,Confidence,and Phenotype Vector.
    CropConfPheno = []
    CropConfTissueComm = []
    TissueArea_Conf = []
    for c in range(clusters[0]):
        CropConfPheno.append([])
            
    for c in range(clusters[1]):
        CropConfTissueComm.append([])  
    
    for c in range(clusters[2]):
        TissueArea_Conf.append([])  

    # Get lists of patches
    for cluster_level in result_images:  
        for clust_idx, cluster in enumerate(cluster_level[0]):
            for patch in cluster:  
                CropConfPheno[clust_idx].append(patch)
        for clust_idx, cluster in enumerate(cluster_level[1]):
            for patch in cluster:  
                CropConfTissueComm[clust_idx].append(patch)
        for r_i, r in enumerate(cluster_level[2]):
            TissueArea_Conf[r_i].append(r)
            
    return CropConfPheno,CropConfTissueComm, TissueArea_Conf

def topk_confident_patches(dataset,n_cell_types,cell_type_top1,
                            cell_type_assignment,cell_type_idx,CropConfPheno,
                            CropConfTissueComm,im,imList,PCL_reprsntions,count):
    '''
    '''
    K = 25

    # Create label map. 
    labels,labels_list = load_patch_image(dataset,imList)
    for ll in range(len(labels_list)):
        labels_list[ll] = np.transpose(labels_list[ll])
    labels = np.concatenate(labels_list,axis=1)

            
    for c in range(n_cell_types):
                                                                        
        # Obtain patch info of the K with most certainty
        highest_ent_patch = cell_type_assignment[np.where(cell_type_top1[:PCL_reprsntions.shape[0]]==c)[0],c].argsort()[-K:]
        for patch_idx in np.where(cell_type_top1[:PCL_reprsntions.shape[0]]==c)[0][highest_ent_patch]:                  
                    
            # Select the patch from the patch_image            
            mask = labels==patch_idx

            if cell_type_idx==0:
                mx_val = mask.argmax(0).max()*dataset.patch_size
                mxx_val = mask.argmax(1).max()*dataset.patch_size
                CropConfPheno[c].append([im[mx_val:mx_val+dataset.patch_size,mxx_val:mxx_val+dataset.patch_size].copy(), # The original Image
                                            cell_type_assignment[patch_idx,c], # The cell type certainty
                                            [], # The parameters obtained from contrastive learning
                                            [100000*count+patch_idx]]) # Number of the image, and patch identificator
            elif cell_type_idx==1:
                mx_val = mask.argmax(0).max()*dataset.patch_size
                mxx_val = mask.argmax(1).max()*dataset.patch_size
                minIdx = mx_val-dataset.patch_size*2
                maxIdx = mx_val+dataset.patch_size*3
                minIdy = mxx_val-dataset.patch_size*2
                maxIdy = mxx_val+dataset.patch_size*3
                if maxIdx>im.shape[0] or maxIdy>im.shape[1] or minIdx<0 or minIdy<0:
                    continue
                CropConfTissueComm[c].append([im[minIdx:maxIdx,minIdy:maxIdy].copy(), # The original Image
                                            cell_type_assignment[patch_idx,c], # The cell type certainty
                                            [100000*count+patch_idx]]) # Patch identificator 
    
    return CropConfPheno,CropConfTissueComm

def save_2Dmatrix_in_excel_with_names(filename,matrix,Names):
    dict_ = {}
    for n, name in enumerate(Names):
        dict_[name] = matrix[:,n]
    dict_ = pd.DataFrame.from_dict(dict_)      
    dict_.to_excel(filename) 

def save_heatmap_with_names(filename,matrix,Names):    
    if matrix.shape[0]>2:
        plt.close()
        plt.figure()
        sns.set(font_scale=1.1)
        h_E_Fig = sns.clustermap(matrix, col_cluster=True, row_cluster=True, xticklabels=Names, linewidths=0,vmin=-2, vmax=2, cmap="bwr")                
        h_E_Fig.savefig(filename,dpi=600) 

def calculate_pearson_correlation(matrix_0,matrix_1):
    cor_val = stats.spearmanr(matrix_0.flatten(),matrix_1.flatten()).correlation
    cor_val = 0 if np.isnan(cor_val) else cor_val
    return cor_val

def calculate_IoU_usingOtsu(matrix_0,matrix_1):
    # Check if all values are equal and return an IoU of zeros
    if matrix_0.max()==matrix_0.min() or matrix_1.max()==matrix_1.min():
        return 0

    matrix_0 = matrix_0>filters.threshold_otsu(matrix_0)
    matrix_1 = matrix_1>filters.threshold_otsu(matrix_1)        
    intersection = np.logical_and(matrix_0, matrix_1)
    union = np.logical_or(matrix_0, matrix_1)
    return np.sum(intersection) / np.sum(union)    

def calculate_marker_colocalization(matrix,MarkerNames):
    # matrix: contains (number of patches, x_dimension, y_dimension, number of markers)
    Marker_Colocalization =  np.zeros(len(list(itertools.combinations(MarkerNames,2))))
    for n_comb, pair_of_markers in enumerate(itertools.combinations(MarkerNames,2)):
        id_0 = MarkerNames.index(pair_of_markers[0])
        id_1 = MarkerNames.index(pair_of_markers[1])
        matrix_0 = matrix[:,:,:,[id_0]]
        matrix_1 = matrix[:,:,:,[id_1]]
        for n_patch in range(matrix.shape[0]):            
            Marker_Colocalization[n_comb] += calculate_pearson_correlation(matrix_0[n_patch,:,:],matrix_1[n_patch,:,:])
    return Marker_Colocalization/matrix.shape[0]

def calculate_Cell_Size_circularity(filename, matrix):
    # matrix: contains (number of patches, x_dimension, y_dimension)         
    # selem = disk(2)
    Size_eccentricity = np.zeros((2))
    for patch in range(matrix.shape[0]):
        if matrix[patch,:,:].sum()>0 and (matrix[patch,:,:].max()-matrix[patch,:,:].min())>0:
            logical = matrix[patch,:,:]>filters.threshold_otsu(matrix[patch,:,:]) 
            # logical = opening(logical, selem)
            Size_eccentricity[0] += logical.sum() # Area
            Size_eccentricity[1] += min((4*3.14159*logical.sum())/(perimeter(logical,neighbourhood=4)**2),1) # Eccentricity
            # if patch<4:
            #     cv2.imwrite(filename+"Patch_"+str(patch)+"_Size_"+str(logical.sum())+"_Ecc_"+str((4*3.14159*logical.sum())/(perimeter(logical,neighbourhood=4)**2))+"_Raw.png", matrix[patch,:,:]/(matrix[patch,:,:].max())*255)
            #     cv2.imwrite(filename+"Patch_"+str(patch)+"_Size_"+str(logical.sum())+"_Ecc_"+str((4*3.14159*logical.sum())/(perimeter(logical,neighbourhood=4)**2))+"_Logical.png", logical*255)
    return Size_eccentricity/matrix.shape[0]

def extract_topk_patches_from_cohort(dataset, CropConf, Marker_Names,cell_type,Thresholds):
    '''
    docstring
    '''

    thisfolder = dataset.bioInsights_dir_cell_types + cell_type+'/'

    if cell_type=='Phenotypes':
        mult_1 = 1 
        mult_2 = 2
    else:
        mult_1 = 5 
        mult_2 = 4

    ## Iterate through Phenotypes to extract topk patches
    k=81
    topkPatches=[]
    # Create a heatmap marker using topk patches
    heatmapMarkerExpression = np.zeros((7,len(CropConf),len(Marker_Names))) #Number of TMEs x number of markers
    heatmap_Colocalization = np.zeros((len(CropConf),int(len(Marker_Names)*(len(Marker_Names)-1)/2))) # Number of TMEs x Number of Marker combinations
    heatmap_CellSize_circularity = np.zeros((len(CropConf),2)) # Number of TMEs x (size and circularity)

    # Use CropCOnf, that saves a lot of patches...
    for n_cell_type ,CropConf_i in enumerate(CropConf):
        
        if len(CropConf_i)==0: # Let te heatmap be zero in case no ells were assigned to this TME.
            continue
        
        # Choose patches with most confidence
        topkPheno = np.array([CCP[1] for CCP in CropConf_i]).argsort()[-k:]        
        # Save topkPheno to heatMarkerMap. Mean
        MarkerExpression = np.array([CropConf_i[t][0] for t in topkPheno])

        heatmap_Colocalization[n_cell_type,:] = calculate_marker_colocalization(MarkerExpression,Marker_Names)
        try:
            heatmap_CellSize_circularity[n_cell_type,:] = calculate_Cell_Size_circularity(thisfolder,MarkerExpression[:,:,:,Marker_Names.index('DAPI')])
        except:
            print('could not calculate cell size circularity because DAPI marker is missing')

        MarkerExpression = np.reshape(MarkerExpression,(MarkerExpression.shape[0]*MarkerExpression.shape[1]*MarkerExpression.shape[2],MarkerExpression.shape[3]))
        heatmapMarkerExpression[0,n_cell_type,:] = np.mean(MarkerExpression,axis=0)                    
        for n_i, i in enumerate(Thresholds[1:]):
            heatmapMarkerExpression[n_i+1,n_cell_type,:] = np.percentile(MarkerExpression,i,axis=0)                    

        # Save Image in RGB
        ImwithKPatches = np.zeros((dataset.patch_size*mult_1*int(np.sqrt(k))+mult_2*int(np.sqrt(k)),dataset.patch_size*mult_1*int(np.sqrt(k))+mult_2*int(np.sqrt(k)),CropConf_i[0][0].shape[2]))
        for t_n, t in enumerate(topkPheno):
            row = np.floor(t_n/int(k**0.5))
            col = np.mod(t_n,int(k**0.5))
       
            # Assign patch to Image
            ImwithKPatches[int(row*mult_1*dataset.patch_size+row*mult_2):int((row+1)*mult_1*dataset.patch_size+row*mult_2),int(col*mult_1*dataset.patch_size+col*mult_2):int((col+1)*mult_1*dataset.patch_size+col*mult_2),:] = CropConf_i[t][0]
        
        if len(Marker_Names)<10:
            # Fill unassigned patches with zeroes.
            for t_n in range(len(topkPheno),k):
                row = np.floor(t_n/int(k**0.5))
                col = np.mod(t_n,int(k**0.5))
                # Assign patch to Image
                ImwithKPatches[int(row*mult_1*dataset.patch_size+row*mult_2):int((row+1)*mult_1*dataset.patch_size+row*mult_2),int(col*mult_1*dataset.patch_size+col*mult_2):int((col+1)*mult_1*dataset.patch_size+col*mult_2),:] = 0.0
                            
            # RGBImwithKPatches = dataset.nPlex2RGB(ImwithKPatches)
            imwrite(thisfolder+'Cell_type_{}_Raw.tiff'.format(n_cell_type+1),np.moveaxis(ImwithKPatches,2,0))
            
            # Save Certainty of this Phenotype
            plt.close()
            plt.figure()
            n, bins, patches = plt.hist(np.array([i[1] for i in CropConf_i]), 100, color=cm.jet_r(int(n_cell_type*(255/int(len(CropConf))))), alpha=1)            
            plt.ylabel('Number of Superpatches',fontsize=16)            
            plt.xlabel('Level of cell type certainty',fontsize=16)
            plt.title('Histogram of phenotype '+str(n_cell_type+1)+' Certainty',fontsize=16)
            plt.savefig(thisfolder+'ConfidenceHistogram_{}.png'.format(n_cell_type+1), format="PNG",dpi=600)                                         

        # Assign topkPheno Patches
        topkPatches+=[CropConf_i[t] for t in topkPheno]     

    return heatmapMarkerExpression, heatmap_Colocalization, heatmap_CellSize_circularity

def save_heatmap_raw_and_normalized(filename, heatmap, TME_names,Colormap,Marker_Names):    
    y_ticklabels = TME_names[heatmap.sum(1)!=0]
    c_map = Colormap[:heatmap.shape[0]][heatmap.sum(1)!=0]
    # Figure Heatmap Raw Values
    plt.close()
    plt.figure()
    sns.set(font_scale=1.1)
    cluster=True if (heatmap.sum(1)!=0).sum()>1 else False
    h_E_Fig = sns.clustermap(heatmap[heatmap.sum(1)!=0,:],col_cluster=cluster, row_cluster=cluster, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(filename+'_Raw.png',dpi=600) 

    # Figure Heatmap Min is 0 and max is 1 Values    
    h_E_COL_MinMax = heatmap[heatmap.sum(1)!=0,:] - heatmap[heatmap.sum(1)!=0,:].min(0,keepdims=True)
    h_E_COL_MinMax = h_E_COL_MinMax/h_E_COL_MinMax.max(0,keepdims=True)
    h_E_COL_MinMax[np.isnan(h_E_COL_MinMax)] = 0 
    plt.close()
    plt.figure()
    sns.set(font_scale=1.1)
    if (h_E_COL_MinMax.sum(1)!=0).sum()==1: # One column heatmap
        h_E_Fig = sns.clustermap(h_E_COL_MinMax[h_E_COL_MinMax.sum(1)!=0,:],col_cluster=False, 
            row_cluster=False, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_MinMax.png',dpi=600) 
    elif (h_E_COL_MinMax.sum(1)!=0).sum()>1: # Normal heatmap.    
        h_E_Fig = sns.clustermap(h_E_COL_MinMax[h_E_COL_MinMax.sum(1)!=0,:],col_cluster=True, 
            row_cluster=True, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_MinMax.png',dpi=600)  
    
    # Figure Heatmap z-scored values
    h_E_COL_Norm = stats.zscore(heatmap[heatmap.sum(1)!=0,:],axis=0)  
    h_E_COL_Norm[np.isnan(h_E_COL_Norm)] = 0              
    plt.close()
    plt.figure()
    sns.set(font_scale=1.1)    
    if (h_E_COL_Norm.sum(1)!=0).sum()==1: # One column heatmap
        h_E_Fig = sns.clustermap(h_E_COL_Norm,col_cluster=False, vmin=-2, vmax=2, row_cluster=False, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_Norm.png',dpi=600) 
    elif (h_E_COL_Norm.sum(1)!=0).sum()>1: # Normal heatmap
        h_E_Fig = sns.clustermap(h_E_COL_Norm,col_cluster=True, vmin=-2, vmax=2, row_cluster=True, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_Norm.png',dpi=600) 

def save_heatMapMarker_and_barplot(dataset, heatmapMarkerExpression, heatmapMarkerColocalization, heatmap_CellSize_circularity, CropConf,Marker_Names,cell_type,Thresholds):
    '''
    '''
    if cell_type=='Phenotypes':
        abrev = 'P'
    else:
        abrev = 'N'

    Colormap=cm.jet(range(0,255,int(255/heatmapMarkerExpression.shape[1])))[:,:3]

    # Save heatmapmarker to disk
    for n_j, j in enumerate(Thresholds):
        save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types+cell_type+'/heatmap_MarkerExpression_'+str(j), 
                                        heatmapMarkerExpression[n_j,:,:], np.array([abrev+str(i+1) for i in range(len(CropConf))]),
                                        Colormap,Marker_Names)
        save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types+cell_type+'/heatmap_MarkerExpression_Colocalization_'+str(j), 
                                        np.concatenate((heatmapMarkerExpression[n_j,:,:],heatmapMarkerColocalization),axis=1), np.array([abrev+str(i+1) for i in range(len(CropConf))]),
                                        Colormap,Marker_Names+['_'.join(i) for i in itertools.combinations(Marker_Names,2)])
        

    try:
        save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types + cell_type + '/heatmap_Size_Circularity', heatmap_CellSize_circularity, np.array([abrev+str(i+1) for i in range(len(CropConf))]),Colormap,['Nucleus Size', 'Nucleus circularity'])
    except:
        print('')
    save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types + cell_type + '/heatmap_Colocalization', heatmapMarkerColocalization, np.array([abrev+str(i+1) for i in range(len(CropConf))]),Colormap,['_'.join(i) for i in itertools.combinations(Marker_Names,2)])

    # Calculate Phenotype abundance across all patients
    plt.close()
    plt.figure()
    sns.set(font_scale=1.0)
    BarPlotPresenceOfPhenotypes = sns.barplot(x=[abrev+str(i+1) for i in range(len(CropConf))] ,y=np.array([len(i) for i in CropConf]),palette=Colormap)      
    BarPlotPresenceOfPhenotypes.set(xlabel=cell_type, ylabel = 'Number of patches',title='Histogram of abundance across the patient cohort')                
    BarPlotPresenceOfPhenotypes.set_xticklabels([abrev+str(i+1) for i in range(len(CropConf))], size=7)
    plt.savefig(dataset.bioInsights_dir_cell_types + cell_type +'/Barplot_cell_types.png',dpi=600)

def neigh_comp(TC,phenoInd):
    InteractivityVect = np.zeros(len(phenoInd))
    for n_Phen, PH in enumerate(phenoInd):
        PH = [p[0] for p in PH]
        if len(PH)>0:
            for t in TC:
                if t[2] in PH:
                    InteractivityVect[n_Phen]+=1     
    return InteractivityVect

def obtain_neighborhood_composition(dataset,CropConfPheno,CropConfTissueComm):
    '''
    '''

    # Find indices of phenotypes
    phenoInd = []
    for c in CropConfPheno:
        # Find all indices
        phenoInd.append([c_n[3] for c_n in c])        

    # Generate Interactivity matrix
    dict_neigh = []
    for n_Neighbor, TC in enumerate(CropConfTissueComm):
        dict_neigh.append({'TC':TC,'phenoInd':phenoInd})
    result = parallel_process(dict_neigh,neigh_comp,use_kwargs=True,front_num=1,desc='BioInsights: Calculate phenotype abundance whithin neighborhoods') 
    InteractivityMatrix = np.stack(result)

    # Input
    Colormap_Pheno=cm.jet_r(range(0,255,int(255/len(CropConfPheno))))[:,:3]
    Colormap_Neigh=cm.jet_r(range(0,255,int(255/len(CropConfTissueComm))))[:,:3]

    # Save interactivity matrix
    sns.set(font_scale=1.5)
    plt.close()
    heatmapInteractivityMatrix_Fig = sns.clustermap(InteractivityMatrix,col_cluster=False, row_cluster=False, row_colors=Colormap_Neigh, col_colors=Colormap_Pheno,xticklabels=['P'+str(i+1) for i in range(len(CropConfPheno))],yticklabels=['N'+str(i+1) for i in range(len(CropConfTissueComm))], linewidths=0.5, cmap="bwr")            
    plt.xlabel("Phenotypes")
    plt.ylabel("Neighborhoods")
    heatmapInteractivityMatrix_Fig.savefig(dataset.bioInsights_dir_cell_types+'Neighborhoods/heatmap_InteractivityMat_Raw.png',dpi=600) 
    plt.close()
    InteractivityMatrix[InteractivityMatrix==0]=1e-3
    heatmapInteractivityMatrix_Fig = sns.clustermap(stats.zscore(InteractivityMatrix,axis=1),col_cluster=False, row_cluster=False, row_colors=Colormap_Neigh, col_colors=Colormap_Pheno,xticklabels=['P'+str(i+1) for i in range(len(CropConfPheno))],yticklabels=['N'+str(i+1) for i in range(len(CropConfTissueComm))], linewidths=0.5, cmap="bwr")            
    plt.xlabel("Phenotypes")
    plt.ylabel("Neighborhoods")
    heatmapInteractivityMatrix_Fig.savefig(dataset.bioInsights_dir_cell_types+'Neighborhoods/heatmap_InteractivityMat_Norm.png',dpi=600) 

def Area_to_Neighborhood_to_Phenotype(dataset,clusters,IndexAndClass, ClusteringThreshold):
    
    # Initialize variables for neighborhood and areas            
    Area_to_Neigh = np.zeros((clusters[-1],clusters[-2]))                
    Neigh_to_Pheno = np.zeros((clusters[-2],clusters[-3]))            

    # Obtain clusters per Slide
    for count, idxclster in enumerate(IndexAndClass):           
        try:
            neigh_to_area_assignment = np.load(osp.join(dataset.processed_dir_cell_types,'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(idxclster[1], clusters[-1])))                
            patch_to_neigh_assignment = np.load(osp.join(dataset.processed_dir_cell_types,'cluster_assignmentPerPatch_Index_{}_0_ClustLvl_{}.npy'.format(idxclster[1], clusters[-2])))        
            patch_to_pheno_assignment = np.load(osp.join(dataset.processed_dir_cell_types,'cluster_assignmentPerPatch_Index_{}_0_ClustLvl_{}.npy'.format(idxclster[1], clusters[-3])))         
            data = dataset.get(idxclster[1],0)
            patch_to_neigh_assignment[data.num_nodes:,:] = 0
            patch_to_pheno_assignment[data.num_nodes:,:] = 0
        except:
            continue

        # Normalize each pheno, neigh, area
        neigh_to_area_assignment = neigh_to_area_assignment-neigh_to_area_assignment.min(0,keepdims=True)
        neigh_to_area_assignment = neigh_to_area_assignment/neigh_to_area_assignment.max(0,keepdims=True)
        patch_to_neigh_assignment = patch_to_neigh_assignment-patch_to_neigh_assignment.min(0,keepdims=True)
        patch_to_neigh_assignment = patch_to_neigh_assignment/patch_to_neigh_assignment.max(0,keepdims=True)
        patch_to_pheno_assignment = patch_to_pheno_assignment-patch_to_pheno_assignment.min(0,keepdims=True)
        patch_to_pheno_assignment = patch_to_pheno_assignment/patch_to_pheno_assignment.max(0,keepdims=True)

        # Assign area to neighborhood
        Area_to_Neigh += np.transpose(neigh_to_area_assignment)
        
        # Calculate percentile        
        PercTHrsl_area = np.percentile(neigh_to_area_assignment[:data.num_nodes,:],axis=0,q=ClusteringThreshold)
        PercTHrsl_neigh = np.percentile(patch_to_neigh_assignment[:data.num_nodes,:],axis=0,q=ClusteringThreshold)
        PercTHrsl_pheno = np.percentile(patch_to_pheno_assignment[:data.num_nodes,:],axis=0,q=ClusteringThreshold)
        
        # Assign neighborhood
        for i in range(neigh_to_area_assignment.shape[1]):
            neigh_to_area_assignment[:,i][PercTHrsl_area[i]>=neigh_to_area_assignment[:,i]] = 0             

        # Assign neighborhood
        for i in range(patch_to_neigh_assignment.shape[1]):
            patch_to_neigh_assignment[:,i][PercTHrsl_neigh[i]>=patch_to_neigh_assignment[:,i]] = 0 
        
        # Assign Phenotype
        for i in range(patch_to_pheno_assignment.shape[1]):
            patch_to_pheno_assignment[:,i][PercTHrsl_pheno[i]>=patch_to_pheno_assignment[:,i]] = 0                 

        Neigh_to_Pheno += np.matmul(np.transpose(patch_to_neigh_assignment[:data.num_nodes,:]),patch_to_pheno_assignment[:data.num_nodes,:])

    # Save heatmap Neigh_to_Pheno Normalized
    plt.close()
    plt.figure()
    sns.set(font_scale=1.1)
    row_colors_Neigh=cm.jet_r(range(0,255,int(255/clusters[1])))[:,:3]
    yticklabels_Neigh = ['N'+str(i+1) for i in range(clusters[1])]
    col_colors_Pheno=cm.jet_r(range(0,255,int(255/clusters[0])))[:,:3]
    xticklabels_Pheno = ['P'+str(i+1) for i in range(clusters[0])]
    Neigh_to_Pheno = Neigh_to_Pheno/Neigh_to_Pheno.sum(1,keepdims=True)*100
    Neigh_to_Pheno[np.isnan(Neigh_to_Pheno)] = 0
    h_E_Fig = sns.clustermap(Neigh_to_Pheno.transpose(),col_cluster=True, row_cluster=True, col_colors=row_colors_Neigh,
                                row_colors=col_colors_Pheno,yticklabels=xticklabels_Pheno,xticklabels=yticklabels_Neigh, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(dataset.bioInsights_dir_cell_types + 'Neighborhoods/heatmap_Phenotype_composition_of_neighborhoods_Perc_Thrs{}.png'.format(str(ClusteringThreshold)),dpi=600) 

    # Save heatmap Neigh_to_Pheno Normalized
    plt.close()
    plt.figure()
    sns.set(font_scale=1.1)
    row_colors_Area=cm.jet_r(range(0,255,int(255/clusters[2])))[:,:3]
    yticklabels_Area = ['A'+str(i+1) for i in range(clusters[2])]
    col_colors_Neigh=cm.jet_r(range(0,255,int(255/clusters[1])))[:,:3]
    xticklabels_Neigh = ['N'+str(i+1) for i in range(clusters[1])]
    Area_to_Neigh = Area_to_Neigh/Area_to_Neigh.sum(1,keepdims=True)*100    
    Area_to_Neigh[np.isnan(Area_to_Neigh)] = 0
    h_E_Fig = sns.clustermap(Area_to_Neigh.transpose(),col_cluster=True, row_cluster=True, col_colors=row_colors_Area,
                                row_colors=col_colors_Neigh,yticklabels=xticklabels_Neigh,xticklabels=yticklabels_Area, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(dataset.bioInsights_dir_cell_types + 'heatmap_Neighborhood_composition_of_Areas_Perc.png',dpi=600) 



import torch
import torch.nn.functional as F
import NaroNet.NaroNet_model.pooling as pooling
import NaroNet.NaroNet_model.loss as loss
import NaroNet.NaroNet_model.GNN as GNN
import NaroNet.utils.utilz
import numpy as np
import copy

class NaroNet_model_simple(torch.nn.Module):
    def __init__(self, num_features, labels, num_nodes, clusts, args):
        super(NaroNet_model_simple, self).__init__()
        self.features = num_features        
        self.args = args                                 

        # Learning local phenotypes 
        self.pheno_clust = GNN.pheno(self.features,clusts[0], args)
                        
        # Learning cellular neighborhoods
        self.neigh_clust = GNN.neigh(self.features, clusts[1], args,True)  
        self.neigh_emb = GNN.neigh(self.features, clusts[1], args, False)            
        
        # Learning tissue areas
        self.area_clust = GNN.area(self.features, clusts[2], args,True)   
        self.area_emb = GNN.area(self.features, clusts[2], args,False)                        

        # Final layer
        self.lin_pat_1 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], len(set(labels[0])), bias=True)

        if len(args['experiment_Label'])>1:
            self.lin_pat_2 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], int(max(labels[1])) if len(labels[1])>5 else len(labels[1]), bias=True)          
        if len(args['experiment_Label'])>2:
            self.lin_pat_3 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], int(max(labels[2])) if len(labels[2])>5 else len(labels[2]), bias=True)          
        if len(args['experiment_Label'])>3:
            self.lin_pat_4 = torch.nn.Linear(clusts[0] + clusts[1] + clusts[2], int(max(labels[3])) if len(labels[3])>5 else len(labels[3]), bias=True)          
    
    def reset_parameters(self):        
                
        # Phenotypes
        self.pheno_clust.reset_parameters()
                
        # Cellular neighborhoods
        self.neigh_clust.reset_parameters()

        # Tissue areas
        self.area_clust.reset_parameters()

        # Final layer    
        self.lin_pat_1.reset_parameters()                
        if len(self.args['experiment_Label'])>1:
            self.lin_pat_2.reset_parameters()
        if len(self.args['experiment_Label'])>2:
            self.lin_pat_3.reset_parameters()
        if len(self.args['experiment_Label'])>3:
            self.lin_pat_4.reset_parameters()
                    
    def softmaxToClst(self,s,args,device):
        s=F.softmax(s,dim=-1)
        if args['1Patch1Cluster']:
            s = torch.where(s>=s.max(-1).values.unsqueeze(-1).repeat(1,1,s.shape[2]), s, torch.tensor([0],dtype=torch.float32).to(device))
        s = torch.where(s>=args['attntnThreshold'], s, s*0)
        return s
    
    def poolingToClst(self,s,device,num_nodes):
        Spatient = torch.zeros(s.shape[0],s.shape[-1],dtype=torch.float32).to(device)
        for i in range(s.shape[0]):
            Spatient[i,:] = (s[i,:num_nodes[i],:].sum(0)/num_nodes[i])   
        return Spatient

    def ObtainPhenotypesClustering(self, x, data, device, args):          
        self.s=[self.pheno_clust(x, data.edge_index, device, data.num_nodes,args)]        
        return self.s

    def ClassifyPatients(self, args):
        x = []
        
        # Fully connected Layer               
        x.append(self.lin_pat_1(torch.cat(self.S,dim=-1)))                         
        if len(self.args['experiment_Label'])>1:
            x.append(self.lin_pat_2(torch.cat(self.S,dim=-1)))
        if len(self.args['experiment_Label'])>2:
            x.append(self.lin_pat_3(torch.cat(self.S,dim=-1)))
        if len(self.args['experiment_Label'])>3:
            x.append(self.lin_pat_4(torch.cat(self.S,dim=-1)))
                
        return x

    def Areas_forward(self, data, device, args):
        
        # GNN Forward to obtain clusters.
        self.s.append(self.area_clust(data.x, data.edge_index, device, data.num_nodes,args))                            
        
        # Obtain Cell Entropy Loss
        contrast_loss_Ar, PatchEntropy_loss_Ar = loss.patch_contrast_entropy(self.area_emb(data.x, data.edge_index, device, data.num_nodes,args),F.softmax(self.s[-1],dim=2), [self.s[-1].shape[1]]*self.s[-1].shape[0],device)

        # Apply Softmax to cluster assignment             
        self.s[-1] = self.softmaxToClst(self.s[-1],args,device)
                
        # Obtain Patient Concentration for each cluster
        self.S.append(self.poolingToClst(self.s[-1],device,[self.s[-1].shape[1]]*self.s[-1].shape[0]))                           
        self.s[-1] = self.s[-1].to('cpu') 
        return data.x, data.edge_index, contrast_loss_Ar, PatchEntropy_loss_Ar

    def Neighborhoods_forward(self, data, device, args):

        # Neighborhood assignment
        self.s.append(self.neigh_clust(data.x, data.edge_index, device, data.num_nodes,args))  

        # Obtain patch
        contrast_loss_Nb, PatchEntropy_loss_Nb = loss.patch_contrast_entropy(self.neigh_emb(data.x, data.edge_index, device, data.num_nodes,args),F.softmax(self.s[-1],dim=2), data.num_nodes,device)
                
        # Apply Softmax to cluster assignment       
        self.s[-1] = self.softmaxToClst(self.s[-1],args,device)

        # Node-Pooling     
        data.x, data.edge_index = pooling.Dense_Pooling(data.x, data.edge_index, self.s[-1], device, data.num_nodes, True)  # Tissue-Communities                                             

        # Obtain Patient Concentration for each cluster
        self.S.append(self.poolingToClst(self.s[-1],device,data.num_nodes)) 
        self.s[-1] = self.s[-1].to('cpu')           

        return data.x, data.edge_index, contrast_loss_Nb, PatchEntropy_loss_Nb

    def Phenotypes_forward(self, data, device, args):
        
        # Assign patches to phenotypes
        self.s = self.ObtainPhenotypesClustering(data.x,data, device, args)                                          

        # Calculate patch entropy and contrast loss
        contrast_loss_Ph, PatchEntropy_loss_Ph = loss.patch_contrast_entropy(data.x,F.softmax(self.s[0],dim=2), data.num_nodes,device)
                        
        # Apply Softmax to cluster assignment
        self.s[-1] = self.softmaxToClst(self.s[-1],args,device)        
        
        # Calculate subject abundance
        self.S = [self.poolingToClst(self.s[-1],device,data.num_nodes)]    
        self.s[-1] = self.s[-1].to('cpu')                   

        return contrast_loss_Ph, PatchEntropy_loss_Ph

    def forward(self, data, device, doClustering,args):
        
        if doClustering:                                
            # Calculate phenotypes            
            contrast_loss_Ph, PatchEntropy_loss_Ph = self.Phenotypes_forward(data, device, args)            
            
            # Calculate neighborhoods                      
            data.x, data.edge_index, contrast_loss_Nb, PatchEntropy_loss_Nb = self.Neighborhoods_forward(data, device, args)
                                  
            return data, contrast_loss_Ph, PatchEntropy_loss_Ph, contrast_loss_Nb, PatchEntropy_loss_Nb, self.s                    
        
        else:        
            # Obtain Tissue-Communities Interactions
            data.x, data.edge_index, contrast_loss_Ar, PatchEntropy_loss_Ar = self.Areas_forward(data, device, args)

        # Classify subjects from microenvironment abundances        
        x = self.ClassifyPatients(args)                       
                
        return x, self.s[2:], contrast_loss_Ar, PatchEntropy_loss_Ar